/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Point on the sphere S^2
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _POINT_S2_H_
#define _POINT_S2_H_

// includes, file


//! Point on the unit sphere S^2
template<class T>
class PointS2 {

public:

  //! Constructor, default (creates a point on the North pole)
  PointS2();

  //! Constructor
  //! @param  t1 / t2 / t3  x / y / z location of point 
  PointS2( const T& t1, const T& t2, const T& t3);

  //! Constructor
  //! @param  theta / phi  location of point in spherical coordinates
  PointS2( const T& theta, const T& phi);

  //! Destructor
  ~PointS2();

  //! Constructor, copy
  PointS2( const PointS2& p);

  //! Assignment operator
  PointS2 operator=( const PointS2& p);

public:

  //! Equality operator
  bool operator==( const PointS2<T>& rhs) const;

  //! Inequality operator
  bool operator!=( const PointS2<T>& rhs) const;

public:

  // working interface

  //! Rotate point 
  //! @param R  rotation matrix
  template<class MatrixType>
  void rotate( const MatrixType& R);

  //! Rotate point using Euler ZYZ angles
  //! @param z / y / z  Euler ZYZ angles
  template<class S>
    void rotate( const S& z1, const S& y, const S& z2);

  //! Perform euclidian dot product in R^3
  //! @param p2 point to perform dot product with
    T dot( const PointS2<T>& p2);

  //! Access x component
  T& x();

  //! Access y component
  T& y();

  //! Access z component
  T& z();

  //! Access x component
  const T& x() const;

  //! Access y component
  const T& y() const;

  //! Access z component
  const T& z() const;

  //! Access theta
  const T theta() const;

  //! Access phi
  const T phi() const;

 protected:
  // data 

  //! Euclidean coordinates
  T  p_x, p_y, p_z;
};

// include, implementation
#include <point_s2.inl>

#endif 

