/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics functionality.
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_COEFFS_H_
#define _SH_COEFFS_H_

// includes, system
#include <vector>

// includes, project
#include <sh_evaluator.h>
#include <point_s2.h>

//! Spherical Harmonics basis function coefficient vector and functionality 
//! which is needed to work with them
template< class T >
class ShCoeffs {

public:

  typedef  std::pair<T,T>  ShGrad;
  typedef  std::vector<T>  VecT;
  typedef  std::vector<VecT>  ShCoeffsVec;
  typedef  std::vector< std::vector< ShGrad > >  ShGradVec;


public:

  //! Constructor, default
  ShCoeffs();

  //! Constructor
  ShCoeffs( const unsigned int max_sh_band);

  //! Destructor
  ~ShCoeffs();

  //! Constructor, copy
  ShCoeffs( const ShCoeffs& ref);

  //! Assignment operator
  ShCoeffs& operator=( const ShCoeffs& ref);

public:

  //! In place addition operator
  ShCoeffs& operator+=( const ShCoeffs& rhs);

  //! In place subtraction operator
  ShCoeffs& operator-=( const ShCoeffs& rhs);

  //! Addition operator
  const ShCoeffs  operator+( const ShCoeffs& rhs);

  //! Subtraction operator
  const ShCoeffs  operator-( const ShCoeffs& rhs);

public:

  // stream out operator
  template<class S>
  friend std::ostream& operator<<( std::ostream &stream, 
                                   const ShCoeffs<S>& sh_coeffs);

public:

  // public interface

  //! Create an uninitialized coefficient vector.
  //! @param  max_band  Maximum band for which memory is allocated
  void  init( const unsigned int max_sh_band);

  //! Compute Spherical Harmonics function values at p for every basis 
  //! function and store them.
  template<class S>
  void computeYBand( const unsigned int band, const PointS2<S>& p);

  //! Compute Spherical Harmonics function values at p for every basis 
  //! function and store them.
  template<class S>
  void computeY( const unsigned int curr_max_band, const PointS2<S>& p);

  //! Compute Spherical Harmonics function values at p for every basis 
  //! function between band_l and band_u and store them.
  template<class S>
  void computeY( const unsigned int band_l, const unsigned int band_u,
                 const PointS2<S>& p);

  //! Compute Spherical Harmonics function values at p for every basis 
  //! Compute Spherical Harmonics function values at p for every basis 
  //! function and store them in the internal representation
  template<class S>
  void computeY( const PointS2<S>& p);

  //! Compute Spherical Harmonics function values and derivatives at p for 
  //! every basis function and store them in the vector.
  template<class S>
  void computeYGradY( const unsigned int curr_max_band, const PointS2<S>& p);

  //! Compute Spherical Harmonics function values and derivatives at p 
  //! for every basis function and store them in the internal representation
  template<class S>
  void computeYGradY( const PointS2<S>& p);


  //! Reconstruct the signal at p when SH basis function coefficients are 
  //! stored.
  //! @param  point where to reconstruct the signal
  template<class S>
  T  reconstruct( const PointS2<S>& p) const;

  //! Reconstruct the part of the signal in a specific band at p where the point
  //! is those used with computeShVals() previously
  //! @note Assumes the SH basis function coefficients are stored.
  //! @param  band  band for which to reconstruct the signal
  T  reconstructBandAtP( const unsigned int band) const;

  //! Reconstruct the signal and its gradient at p.
  //! @note Precondition: the SH basis function coefficients are stored.
  //! @param  point where to reconstruct the signal
  template<class S>
  T  reconstructYGradY( const PointS2<S>& p) const;

  //! Reconstruct the part of the signal in a specific band and its gradient 
  //! at p where the point is those used with computeShVals() previously
  //! @note Precondition: the SH basis function coefficients are stored.
  //! @param  band  band for which to reconstruct the signal
  //! @param  fval  reconstructed function value for band 
  //! @param  grad  reconstructed gradient for band
  void  reconstructYGradYBandAtP( const unsigned int band, T& f_band, 
                                  ShGrad& grad_band) const;

  //! Create the basis function coefficients corresponding to an acoustic source
  //! @param  loc  location of the acoustic source
  template<class S>
  void  createAcousticSource( const PointS2<S>& loc);

  //! Create the basis function coefficients corresponding to an acoustic source
  //! @param  loc  location of the acoustic source
  template<class S>
  void  createAcousticSource( const unsigned int curr_max_band, 
                               const PointS2<S>& loc);

public:

  //! l_p norm of the coefficient vector
  //! @note By default the l_2 norm is computed
  T  norm( const unsigned int l_p, const signed int band) const;

  //! Helper for l_{\infty} norm
  //! @note Use as sh_coeffs.norm( ShCoeffs<T>::infty());
  static unsigned int 
  infty() {
    return std::numeric_limits<unsigned int>::infinity();
  }

public:

  //! Get a specific coefficient
  const T&  operator()( const unsigned int band, const unsigned int i) const;

  //! Set a specific coefficient
  T&  operator()( const unsigned int band, const unsigned int i);

  //! Get a pointer to the coefficients at a specific band.
  T* getAddress( const unsigned int band);

  //! Get a pointer to the coefficients at a specific band.
  const T* getAddress( const unsigned int band) const;

  //! Get the maximum band for which the coefficient vector is initialized
  unsigned int getMaxBand() const;

  //! Set all elements to a specific value
  void  set( const T& val);

private:

  //! Maximum band for which the coefficient vector is initialized
  unsigned int max_band;

  //! Basis function coefficients for a signal
  ShCoeffsVec  coeffs;

  //! Spherical Harmonics function values (at a specific point)
  //! @note Mostly used as temporary
  ShCoeffsVec  sh_fvals;

  //! Spherical Harmonics gradient function values (at a specific point)
  //! @note Mostly used as temporary
  ShGradVec  sh_grads;

  //! Spherical Harmonics functionality
  //! Spherical Harmonics functionality
  ShEvaluator<T>*  sh_evaluator;

};

// include implementation
#include <sh_coeffs.inl>

#endif // _SH_COEFFS_H_

