/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using RKHS
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_BASE_H_
#define _SH_ROTATION_BASE_H_

// includes, files
#include <sh_coeffs.h>

//! Base class for rotators
template<class T>
class ShRotationBase {

public:

  //! Constructor, standard
  ShRotationBase() { }

  //! Destructor
  virtual ~ShRotationBase() { }

public:

  //! Initialize
  virtual void preload( const unsigned int max_band) = 0;

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  virtual  T 
  rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                   ShCoeffs<T>& sh_coeffs_rotated, 
                   const unsigned int max_band,
                   const T& t1, const T& t2, const T& t3,
                   const unsigned int num_iterations = 1) = 0;

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  //! @param  os_factor  oversampling factor (where applicable)
  virtual  T 
  rotate( const ShCoeffs<T>& /*sh_coeffs*/, 
          ShCoeffs<T>& /*sh_coeffs_rotated*/, 
          const unsigned int /*max_band*/,
          const T& /*t1*/, const T& /*t2*/, const T& /*t3*/,
          const unsigned int /*num_iterations = 1*/) = 0;

  //! Get the name of the rotation algorithm
  virtual const std::string& name() const = 0;
 
  //! Write data out in Matlab format for validation purposed 
  void
  validateRotationMatlab( const unsigned int max_band,
                          const ShCoeffs<T>& sh_coeffs, 
                          const ShCoeffs<T>& sh_coeffs_rotated, 
                          const T& t1, const T& t2, const T& t3)
  {
    std::string fname( "validation_rotation_");
    fname += name();
    fname += ".m";

    std::ofstream of( fname.c_str());

    of << std::scientific;
    of << std::setprecision( 20);

    of << "max_band = " << max_band << ";\n";

    of << "angles = [";
    of << t1 << " " << t2 << " " << t3;
    of << "];\n";

    of << "sh_coeffs = [";
    of << sh_coeffs;
    of << "];\n";

    of << "sh_coeffs_rotated = [";
    of << sh_coeffs_rotated;
    of << "];";

    of.close();
  }



};

#endif 

