/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Evaluation of Spherical Harmonics rotation
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_BENCHMARK_H_
#define _SH_ROTATION_BENCHMARK_H_

// includes, system
#include <vector>

// includes, project
#include <sh_coeffs.h>
#include <rotation.h>
#include <sh_rotation_base.h>

template< class T >
class ShRotationBenchmark {

public:

  //! Constructor, default
  ShRotationBenchmark();

  //! Destructor
  ~ShRotationBenchmark();

public:

  //! Run experiment. Timing experiments are run for signals with 1 up to 
  //! max_band bands. Accuracy experiments are performed for a signal with 
  //! max_band bands and the error is reported per band.
  //! @param  max_band  Maximum band of signals.
  //! @param  num_rand_rotations   Number of random rotations in experiment
  //! @param  sh_coeffs  Spherical Harmonics coefficients of the input signal
  //! @param  num_iterations_timing  Number of iterations for timing
  //!                                measurements
  void runExperiment( const unsigned int max_band, 
                      const unsigned int num_rand_rotations,
                      const unsigned int num_iterations_timing,
                      const ShCoeffs<T>& sh_coeffs );

  //! Run experiment. Timing experiments are run for signals with 1 up to 
  //! max_band bands. Accuracy experiments are performed for a signal with 
  //! max_band bands and the error is reported per band.
  //! @param  max_band  Maximum band of signals.
  //! @param  num_rand_rotations   Number of random rotations in experiment
  //! @param  num_iterations_timing  Number of iterations for timing 
  //!         measurements
  void runExperiment( const unsigned int max_band, 
                      const unsigned int num_rand_rotations,
                      const unsigned int num_iterations_timing);



  //! Run experiment to calculate accuracy. Experiments are run for signals
  //! with 1 up to 
  //! max_band bands.  Rotation angles are a deterministic quasi random sequence
  //! @param  max_band  Maximum band of signals.
  //! @param  num_rand_rotations   Number of random rotations in experiment
  //! @param  offset     Offset into quasi-random rotation sequence at which to 
  //!                    begin
  void runExperimentAccuracy( const unsigned int max_band, 
                              const unsigned int num_rand_rotations,
                    		      const unsigned int offset=0);


public:

  //! Write results to a Matlab file
  void writeResultsMatlab() const;

  //! Write results of accuracy test to a Matlab file
  void writeResultsMatlabAccuracy() const;

private:

  //! Add a new rotator
  template<class S>
  void addRotator( S rotator);

private:

  // Rotators
  std::vector< ShRotationBase<T>* >  rotators;

  // Execution times
  std::vector< std::vector<T> > exec_times;

  // Error measures
  typedef std::vector< std::vector<T> >  VecVec;
  VecVec  errs_avg_l1;
  VecVec  errs_avg_l2;
  VecVec  errs_avg_linfty;
  VecVec  errs_max_l1;
  VecVec  errs_max_l2;
  VecVec  errs_max_linfty;

  //! Rotations, for accuracy experiments
  VecVec rotations_accuracy;
  //! Complete errors for each rotation for accuracy experiments
  std::vector< std::vector< VecVec > >  errs_accuracy;

  //! Generator for random rotations
  Rotation::RandomRotationGenerator<T>  rotation_generator;
};

// include implementation
#include <sh_rotation_benchmark.inl>

#endif // _SH_ROTATION_BENCHMARK_H_

