/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : {lessig|tyler}@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using the algorithm (and code)
///               by Joe Ivanic.
///
///  license    : See README for details. The rotation code used for this method
///               has been obtained from Leslie Greengard.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_IVANIC_H_
#define _SH_ROTATION_IVANIC_H_

// includes, files
#include <sh_rotation_base.h>

// includes, system
#include <string>

// includes, libs
#include <cutil.h>

// includes, project
#include <sh_coeffs.h>

//! Spherical Harmonics rotation as proposed by Greengard and Gimbutas
template<class T>
class ShRotationIvanic : public ShRotationBase<T> {

public:

  //! Constructor, default
  ShRotationIvanic();

  //! Destructor
  ~ShRotationIvanic();

public:

  //! Initialize rotation, in particular load precomputed data
  //! @param  max_band  maximum band to pre-load
  void preload( const unsigned int max_band);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                     ShCoeffs<T>& sh_coeffs_rotated, 
                     const unsigned int max_band,
                     const T& t1, const T& t2, const T& t3,
                     const unsigned int num_iterations = 1);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotate( const ShCoeffs<T>& sh_coeffs, 
            ShCoeffs<T>& sh_coeffs_rotated, 
            const unsigned int max_band,
            const T& t1, const T& t2, const T& t3,
            const unsigned int num_iterations = 1);
  
  //! Get the name of the rotation algorithm
  const std::string& name() const;
  
private:

  //! Constructor, copy
  //! @note Not implemented
  ShRotationIvanic( const ShRotationIvanic&);

  //! Addignment operator
  //! @note Not implemented
  ShRotationIvanic operator=( const ShRotationIvanic&);

private:

  //! Timer for performance measurements
  unsigned int timer;

  static const std::string class_name;

};

// include implementation
#include <sh_rotation_ivanic.inl>

#endif 

