/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation following the paper by Pinchon
///               and Hoggan (Rotation matrices for real spherical harmonics:
///               general rotations of atomic orbitals in space-fixed axes). 
///               Updated code provided by the authors July 2011.
///
///  license    : See README for details. Parts of the code have been obtained 
///               from Didier Pinchon.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_PINCHON_3_H_
#define _SH_ROTATION_PINCHON_3_H_

// includes, files
#include <sh_rotation_base.h>

// includes, system
#include <string>

// includes, libs
#include <cutil.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>

// includes, project
#include <sh_coeffs.h>
#include <pinchon/rotation.h>

//! Spherical Harmonics rotation as proposed by Pinchon and Hoggan
//! @note This class is a wrapper for the code provided by Didier Pinchon and 
//!       we try to use as much of the original functionality as possible.
template<class T>
class ShRotationPinchon3 : public ShRotationBase<T> {

public:

  //! Constructor, default
  ShRotationPinchon3();

  //! Destructor
  ~ShRotationPinchon3();

public:

  //! Initialize rotation, in particular load precomputed data
  //! @param  max_band  maximum band to pre-load
  void preload( const unsigned int max_band);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                     ShCoeffs<T>& sh_coeffs_rotated, 
                     const unsigned int max_band,
                     const T& t1, const T& t2, const T& t3,
                     const unsigned int num_iterations = 1);

  //! Rotate a signal represented in Spherical Harmonics. Assumes
  //! matrices are not preloaded.
  //! @return  Computation time for function values (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3  Euler ZYZ rotation angles
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotate( const ShCoeffs<T>& sh_coeffs,
            ShCoeffs<T>& sh_coeffs_rotated,
            const unsigned int max_band,
            const T& t1, const T& t2, const T& t3,
            const unsigned int num_iterations = 1);

  //! Get the name of the rotation algorithm
  const std::string& name() const;
  
private:

  //! Constructor, copy
  //! @note Not implemented
  ShRotationPinchon3( const ShRotationPinchon3 &);

  //! Addignment operator
  //! @note Not implemented
  ShRotationPinchon3 operator=( const ShRotationPinchon3 &);

private:

  //! Precomputed and preloaded j matrices
  MatL*  mat_j_l;

  //! unrotated vector
  VecL*  sh_vec;
  //! rotated vector
  VecL*  sh_vec_rot;

private:

  //! Timer for performance measurements
  unsigned int timer;

  static const std::string class_name;
};

// include implementation
#include <sh_rotation_pinchon_3.inl>

#endif 

