/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : {lessig|tyler}@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using the algorithm of Greengard 
///               and Gimbutas.
///
///  license    : See README for details. 
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SH_ROTATION_SPECTRAL_H_
#define _SH_ROTATION_SPECTRAL_H_

// includes, files
#include <sh_rotation_base.h>

// includes, system
#include <string>

// includes, libs
#include <cutil.h>

// includes, project
#include <sh_coeffs.h>

//! Spherical Harmonics rotation as proposed by Greengard and Gimbutas
template<class T>
class ShRotationSpectral : public ShRotationBase<T> {

  // MTL matrix and vector convenience typedef's
  typedef  mtl::matrix::dense2D<FType>  Matrix;
  typedef  mtl::dense_vector<FType>  Vector;
  typedef  std::pair<T,T>  ShGrad;

public:

  //! Constructor, default
  ShRotationSpectral();

  //! Destructor
  ~ShRotationSpectral();

public:

  //! Initialize rotation, in particular load precomputed data
  //! @param  max_band  maximum band to pre-load
  //! @param  fname  file name containing the data needed for rotation
  //!                the default name is 'data/sh_rotation_rkhs.dat'
  void preload( const unsigned int max_band);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                     ShCoeffs<T>& sh_coeffs_rotated, 
                     const unsigned int max_band,
                     const T& t1, const T& t2, const T& t3,
                     const unsigned int num_iterations = 1);

  //! Rotate a signal represented in Spherical Harmonics
  //! @return  Computation time of rotation matrix (ms).
  //! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
  //! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
  //!                            (after rotation)
  //! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
  //!                    \a sh_coeffs has to contain at least max_band bands)
  //! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
  //! @param  num_iterations  Number of iterations (for computing rotation
  //!                         matrices and measuring execution time)
  T rotate( const ShCoeffs<T>& sh_coeffs, 
            ShCoeffs<T>& sh_coeffs_rotated, 
            const unsigned int max_band,
            const T& t1, const T& t2, const T& t3,
            const unsigned int num_iterations = 1);
  
  //! Get the name of the rotation algorithm
  const std::string& name() const;
 
private:

  //! Rotate SH signal around the Z axis
  //! @param  sh_coeffs  SH coefficients of unrotated input signal
  //! @param  sh_coeffs_rotated  SH_coefficients of rotated output signal
  //! @param  max_band  maximum band of the SH signal
  //! @param  beta   rotation angle round the Z axis
  //! @param  num_iterations  number of iterations (for improved timing accuracy)
  T  rotateZ( const ShCoeffs<T>& sh_coeffs,
              ShCoeffs<T>& sh_coeffs_rotated,
              const unsigned int max_band,
              const T& beta, 
              const unsigned int num_iterations );


  //! Rotate SH signal around the Y axis
  //! @param  sh_coeffs  SH coefficients of unrotated input signal
  //! @param  sh_coeffs_rotated  SH_coefficients of rotated output signal
  //! @param  max_band  maximum band of the SH signal
  //! @param  beta   rotation angle round the Y axis
  //! @param  num_iterations  number of iterations (for improved timing accuracy)
  T  rotateY( ShCoeffs<T>& sh_coeffs,
              ShCoeffs<T>& sh_coeffs_rotated,
              const unsigned int max_band,
              const T& beta, 
              const unsigned int num_iterations );

  //! Compute sample values for the function and its derivative
  //! @param max_band  maximum band
  //! @param pts       sample points (rotated)
  //! @param beta      rotation angle (a 
  void  computeSampleValuesRotationY( ShCoeffs<T>& sh_coeffs,
                                      const unsigned int max_band, 
                                      std::vector< PointS2<T> >& pts,
                                      const T& beta );

private:

  //! Constructor, copy
  //! @note Not implemented
  ShRotationSpectral( const ShRotationSpectral&);

  //! Addignment operator
  //! @note Not implemented
  ShRotationSpectral operator=( const ShRotationSpectral&);

private:

  //! Temporary storage for the function values
  std::vector< std::vector<T> >  f_vals;

  //! Temporary storage for projected gradients (projected onto rotated theta)
  std::vector< std::vector<T> >  grad_proj;

  //! Timer for performance measurements
  unsigned int timer;

  static const std::string class_name;

};

// include implementation
#include <sh_rotation_spectral.inl>

#endif 

