/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2008      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : util
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Utility functionality for spectral rotation algorithms.
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _SPECTRAL_UTIL_H_
#define _SPECTRAL_UTIL_H_

// includes, system
#include <vector>

// includes, project
#include <point_s2.h>

namespace Spectral {

////////////////////////////////////////////////////////////////////////////////
//! Create \a num equally spaced points on an equi-latitude ring and rotate 
//! them by R
//! @param  num  Number of points to generate
//! @param  theta  latitude location of points (angle measured from the 
//!                positive Z axis, physics convention)
//! @param  pts  Points on the equator
//! @param  R    rotation matrix
////////////////////////////////////////////////////////////////////////////////
template<class T, class M>
void
createPointsEquiLatitudeR( const unsigned int num, const T& theta,
                           std::vector< PointS2<T> >& pts, const M& R) 
{
  T delta = (2.0*PI) / num;
  T phi = 0.0;

  pts.clear();

  for( unsigned int i = 0; i < num; ++i) {

    // create point and rotate it
    PointS2<T> temp( theta, phi);
    temp.rotate( R);
    pts.push_back( temp);

    phi += delta;
  }
} 

//////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics coefficient for m >= 0
//////////////////////////////////////////////////////////////////////////////
template<class T, class S>
T
valCos( const unsigned int i, const S& f_vals_dft, 
        const std::vector<T>& p_nm, const T& scale_p_nm) 
{
  T val_cos = f_vals_dft[i][0] / (p_nm[i] * scale_p_nm);
  val_cos *= (0 == i) ? 0.5 : 1.0;
  return val_cos;
}

//////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics coefficient for m >= 0
//////////////////////////////////////////////////////////////////////////////
template<class T, class S>
T
valCos( const unsigned int i, const S& f_vals_dft, 
        const std::vector<T>& p_nm_scaled ) 
{
  T val_cos = f_vals_dft[i][0] / p_nm_scaled[i];
  return val_cos;
}

//////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics coefficient for m < 0
//////////////////////////////////////////////////////////////////////////////
template<class T, class S>
T
valSin( const unsigned int i, const S& f_vals_dft, 
    const std::vector<T>& p_nm, const T& scale_p_nm) 
{
  return (-1.0 * f_vals_dft[i][1] / (p_nm[i] * scale_p_nm));
}

//////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics coefficient for m < 0
//////////////////////////////////////////////////////////////////////////////
template<class T, class S>
T
valSin( const unsigned int i, const S& f_vals_dft, 
        const std::vector<T>& p_nm_scaled) 
{
  return (-1.0 * f_vals_dft[i][1] / p_nm_scaled[i]);
}

} // end namespace Spectral

#endif // #ifndef _SPECTRAL_UTIL_H_
