/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : rotation
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Rotation utility functionality
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system


// includes, libs
#include <halton.h>
#include <boost/numeric/mtl/mtl.hpp>

namespace Rotation {

////////////////////////////////////////////////////////////////////////////////
//! Construct rotation matrix for ZYZ Euler angles
//! @param t1 / t2 / t3   Euler angles (in rad)
////////////////////////////////////////////////////////////////////////////////
template<class T, class MatrixType>
void
getRotationMatrixEulerZYZ( const T& t1, const T& t2, const T& t3, MatrixType& R)
{  
  // precompute sine's and cosine's of angles
  T  c_1 = cos( t1);
  T  s_1 = sin( t1);

  T  c_2 = cos( t2);
  T  s_2 = sin( t2);

  T  c_3 = cos( t3);
  T  s_3 = sin( t3);

  // assemble matrix
  R(0,0) = c_1 * c_2 * c_3 - s_1 * s_3;
  R(0,1) = - c_2 * c_3 * s_1 - c_1 * s_3;
  R(0,2) = c_3 * s_2;
  R(1,0) = c_3 * s_1 + c_1 * c_2 * s_3;
  R(1,1) = c_1 * c_3 - c_2 * s_1 * s_3;
  R(1,2) = s_2 * s_3;
  R(2,0) = -c_1 * s_2;
  R(2,1) = s_1 * s_2;
  R(2,2) = c_2;
}

////////////////////////////////////////////////////////////////////////////////
//! Construct rotation matrix for Axis angle
//! @param axis1,axis2,axis3, angle  normalized vector and angle in rad
//! @param R  rotation matrix
////////////////////////////////////////////////////////////////////////////////
template<class T, class MatrixType>
void
getRotationMatrixAxisAngle( const T& x, const T& y, const T& z, const T& angle, MatrixType& R)
{  
  // precompute sine's and cosine's of angles
  T  sin_a = sin(angle);
  T  cos_a = cos(angle);

  // assemble matrix
  R(0,0) = cos_a + (1-cos_a) * x * x;
  R(0,1) = (1-cos_a) * x * y - sin_a * z;
  R(0,2) = (1-cos_a) * x * z + sin_a * y;
  R(1,0) = (1-cos_a) * x * y + sin_a * z;
  R(1,1) = cos_a + (1-cos_a) * y * y;
  R(1,2) = (1-cos_a) * y * z - sin_a * x;
  R(2,0) = (1-cos_a) * x * z - sin_a * y;
  R(2,1) = (1-cos_a) * y * z + sin_a * x;
  R(2,2) = cos_a + (1-cos_a) * z * z;
}

////////////////////////////////////////////////////////////////////////////////
//! Extract Euler ZYZ angles from rotation matrix
//! @param R  rotation matrix
//! @param t1 / t2 / t3   Euler angles (in rad)
////////////////////////////////////////////////////////////////////////////////
template<class T, class MatrixType>
void
getRotationEulerZYZMatrix( const MatrixType& R, T& t1, T& t2, T& t3 ) {
  
  t1 = PI/2.0 + atan2( R(2,0), R(2,1));
  t2 = acos( R(2,2));
  t3 = PI/2.0 - atan2( R(0,2), R(1,2));
}

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
RandomRotationGenerator<T>::RandomRotationGenerator() {

  // initalization

  // we require three random numbers
  halton_dim_num_set( 3);

  // skip the first element of the sequence (which is identical zero)
  double vals[3];
  halton( vals);

  // generate random offset
  srand( time( NULL));

  // uncomment the following to generate a random offset
  // unsigned int offset = rand() / 1000000;
  unsigned int offset = 5783;
  for( unsigned int i = 0; i < offset; ++i) {
    halton( vals);
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
RandomRotationGenerator<T>::~RandomRotationGenerator() { }

////////////////////////////////////////////////////////////////////////////////
//! Get the next random rotation as Euler ZYZ rotation
//! @param t1 / t2 / t3  Euler ZYZ rotation angles
////////////////////////////////////////////////////////////////////////////////
template<class T>
void  
RandomRotationGenerator<T>::get( T& t1, T& t2, T& t3) const {

  mtl::matrix::dense2D<T>  M(3,3);
  get( M);

  getRotationEulerZYZMatrix( M, t1, t2, t3);
}

////////////////////////////////////////////////////////////////////////////////
//! Get the next random rotation
//! @note MatrixType has to be a matrix type with mat(i,j) implemented
//! @param mat  rotation matrix
//! @note The implementation is based a memo by Jim Arvo 
//!       (www.ics.uci.edu/~arvo/papers/RotationMat.ps)
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class MatrixType>
void
RandomRotationGenerator<T>::get( MatrixType& mat) const {

  // get three random floats
  double  r_vals[3];
  halton( r_vals);

  // rotation matrix around Z axis
  mtl::matrix::dense2D<T>  R(3,3);
  R(0,0) = cos( 2*PI*r_vals[0]);
  R(0,1) = sin( 2*PI*r_vals[0]);
  R(0,2) = 0.0;
  R(1,0) = -R(0,1);
  R(1,1) = R(0,0);
  R(1,2) = 0.0;
  R(2,0) = 0.0;
  R(2,1) = 0.0;
  R(2,2) = 1.0;

  // vector for Householder reflection
  T  v[3];
  v[0] = cos(2*PI*r_vals[1]) * sqrt(r_vals[2]);
  v[1] = sin(2*PI*r_vals[1]) * sqrt(r_vals[2]);
  v[2] = sqrt(1.0 - r_vals[2]);

  // H = I = 2 * v * v^T
  mtl::matrix::dense2D<T>  H(3,3);
  for( unsigned int i = 0; i < 3; ++i) {
    for( unsigned int j = 0; j < 3; ++j) {
      H(i,j) = 2.0 * v[i] * v[j];
    }
    // subtract identity from diagonal elements
    H(i,i) = -(1.0 - H(i,i));
  }

  mtl::matrix::dense2D<T>  M(3,3);
  mult( H, R, M);

  // copy result to M
  for( unsigned int i = 0; i < 3; ++i) {
    for( unsigned int j = 0; j < 3; ++j) {
      mat(i,j) = M(i,j);
    }
  }
}

} // end namespace Rotation

