/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics functionality.
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, project
#include <util.h>

// includes, system
#include <limits>

// includes, libs
#include <gsl_sf_bessel.h>

#define SH_COEFFS_DEBUG_ASSERT( cond, func_name, msg)                \
  DEBUG_ASSERT( cond, "ShCoeffs", func_name, msg)                  

// namespace, unnamed
namespace {

  //! Dot product between two STL vectors of the same size
  template< class T>
  T
  dot( const std::vector<T>& lhs, const std::vector<T>& rhs) {
    
    SH_COEFFS_DEBUG_ASSERT( lhs.size() == rhs.size(), "dot()", "Size mismatch.");

    T val = 0.0;
    for( unsigned int i = 0; i < lhs.size(); ++i) {
      val += lhs[i] * rhs[i];    
    }

    return val;
  }

  //! Dot product for std::pair
  template<class T>
  std::pair<T,T>
  dot( const std::vector<T>& lhs, const std::vector< std::pair<T,T> >& rhs) {

    SH_COEFFS_DEBUG_ASSERT( lhs.size() == rhs.size(), "dot()", "Size mismatch.");

    std::pair<T,T> val( 0,0);
    for( unsigned int i = 0; i < lhs.size(); ++i) {
      val.first += lhs[i] * rhs[i].first;
      val.second += lhs[i] * rhs[i].second;
    }

    return val;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template< class T >
ShCoeffs<T>::ShCoeffs() :
  max_band( 0),
  coeffs(),
  sh_fvals(),
  sh_grads(),
  sh_evaluator( ShEvaluator<T>::instance())
{ }

////////////////////////////////////////////////////////////////////////////////
//! Constructor
////////////////////////////////////////////////////////////////////////////////
template< class T >
ShCoeffs<T>::ShCoeffs( const unsigned int max_sh_band) :
  max_band( 0),
  coeffs(),
  sh_fvals(),
  sh_grads(),
  sh_evaluator( ShEvaluator<T>::instance())
{
  init( max_sh_band);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template< class T >
ShCoeffs<T>::~ShCoeffs() { }

////////////////////////////////////////////////////////////////////////////////
//! Constructor, copy
////////////////////////////////////////////////////////////////////////////////
template< class T >
ShCoeffs<T>::ShCoeffs( const ShCoeffs& ref) :
  max_band( ref.max_band),
  coeffs( ref.coeffs),
  sh_fvals( ref.sh_fvals),
  sh_grads( ref.sh_grads),
  sh_evaluator( ref.sh_evaluator)
{ }

////////////////////////////////////////////////////////////////////////////////
//! Assignment operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShCoeffs<T>& 
ShCoeffs<T>::operator=( const ShCoeffs& ref) {

  if( this != &ref) {
    max_band = ref.max_band;
    coeffs = ref.coeffs;
    sh_fvals = ref.sh_fvals;
    sh_grads = ref.sh_grads;
    sh_evaluator = ref.sh_evaluator;
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////
//! In place addition operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShCoeffs<T>& 
ShCoeffs<T>::operator+=( const ShCoeffs& rhs) {

  SH_COEFFS_DEBUG_ASSERT( max_band == rhs.max_band, "operator+=", 
                 "max_band for right hand side operand does not match.");

  for( unsigned int band = 0; band < coeffs.size(); ++band) {
    for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
      coeffs[band][i] += rhs.coeffs[band][i];     
    }
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////
//! In place subtraction operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShCoeffs<T>& 
ShCoeffs<T>::operator-=( const ShCoeffs& rhs) {

  SH_COEFFS_DEBUG_ASSERT( max_band == rhs.max_band, "operator+=", 
                 "max_band for right hand side operand does not match.");

  for( unsigned int band = 0; band < coeffs.size(); ++band) {
    for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
      coeffs[band][i] -= rhs.coeffs[band][i];     
    }
  }

  return *this;
}

////////////////////////////////////////////////////////////////////////////////
//! Addition operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
const ShCoeffs<T>  
ShCoeffs<T>::operator+( const ShCoeffs& rhs) {
  return ShCoeffs(*this) += rhs;
}

////////////////////////////////////////////////////////////////////////////////
//! Subtraction operator
////////////////////////////////////////////////////////////////////////////////
template<class T>
const ShCoeffs<T>  
ShCoeffs<T>::operator-( const ShCoeffs& rhs) {
  return ShCoeffs(*this) -= rhs;
}

////////////////////////////////////////////////////////////////////////////////
// stream out operator
////////////////////////////////////////////////////////////////////////////////
template<class S>
std::ostream& 
operator<<( std::ostream &stream, const ShCoeffs<S>& sh_coeffs) {

  for( unsigned int band = 0; band < sh_coeffs.coeffs.size(); ++band) {
    for( unsigned int i = 0; i < sh_coeffs.coeffs[band].size(); ++i) {
      stream << sh_coeffs.coeffs[band][i] << "  ";
    }
    stream << "/  ";
  }

  return stream;
}

////////////////////////////////////////////////////////////////////////////////
//! Create an uninitialized coefficient vector.
//! @param  max_band  Maximum band for which memory is allocated
////////////////////////////////////////////////////////////////////////////////
template<class T>
void  
ShCoeffs<T>::init( const unsigned int max_sh_band) {

  max_band = max_sh_band;

  // allocate max_band+1 because max_band is inclusive
  coeffs.resize( max_band+1);
  sh_fvals.resize( max_band+1);
  sh_grads.resize( max_band+1);

  // allocate the coefficient vectors for the individual bands
  for( unsigned int band = 0; band < coeffs.size(); ++band) {
    coeffs[band].resize( 2*band+1);
    sh_fvals[band].resize( 2*band+1);
    sh_grads[band].resize( 2*band+1);
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics function values at p for every basis 
//! function in the specified band
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
ShCoeffs<T>::computeYBand( const unsigned int band, const PointS2<S>& p) {

  SH_COEFFS_DEBUG_ASSERT( (band <= max_band), "computeY()", "Invalid band.");

  sh_evaluator->shEvalBand( band, p.theta(), p.phi(), sh_fvals[band]);
}

////////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics function values at p for every basis 
//! function and store them internally.
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
ShCoeffs<T>::computeY( const unsigned int c_max_band, const PointS2<S>& p) {

  SH_COEFFS_DEBUG_ASSERT( (c_max_band <= max_band), "computeY()", 
                                                    "Invalid band.");

  sh_evaluator->shEvalBands( c_max_band, p.theta(), p.phi(), sh_fvals);
}

////////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics function values at p for every basis 
//! function between band_l and band_u and store them.
////// //////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
ShCoeffs<T>::computeY( const unsigned int band_l, const unsigned int band_u,
                       const PointS2<S>& p) 
{
  SH_COEFFS_DEBUG_ASSERT( (band_u <= max_band), "computeY()", "Invalid band.");
  SH_COEFFS_DEBUG_ASSERT( (band_l <= band_u), "computeY()", 
                                              "Invalid order of bands.");

  sh_evaluator->shEvalBands( band_l, band_u, p.theta(), p.phi(), sh_fvals);
}

////////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics function values and derivatives at p for 
//! every basis function and store them internally.
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
ShCoeffs<T>::computeYGradY( const unsigned int c_max_band, const PointS2<S>& p) {

  SH_COEFFS_DEBUG_ASSERT( (c_max_band <= max_band), "computeYGradY()", 
                                                    "Invalid band.");

  sh_evaluator->shEvalBands(c_max_band, p.theta(), p.phi(), sh_fvals, sh_grads);
}

////////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics function values at p for every basis 
//! function and store them in the vector.
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
ShCoeffs<T>::computeY( const PointS2<S>& p) {

  computeY( max_band, p);
}

////////////////////////////////////////////////////////////////////////////////
//! Compute Spherical Harmonics function values and derivatives at p for every
//! basis function and store them in the vector.
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
void 
ShCoeffs<T>::computeYGradY( const PointS2<S>& p) {

  computeYGradY( max_band, p);
}

////////////////////////////////////////////////////////////////////////////////
//! Reconstruct the part of the signal in a specific band at p where the point
//! is those used with computeY() previously
//! @note Assumes the SH basis function coefficients are stored.
//! @param  band  band for which to reconstruct the signal
//! @param  point where to reconstruct the signal
////////////////////////////////////////////////////////////////////////////////
template<class T>
T  
ShCoeffs<T>::reconstructBandAtP( const unsigned int band) const {

  SH_COEFFS_DEBUG_ASSERT( band <= max_band, "reconstructBandAtP()",
                                            "Invalid band.");

  return dot( coeffs[band], sh_fvals[band]);
}

////////////////////////////////////////////////////////////////////////////////
//! Reconstruct the signal at p
//! @param  point where to reconstruct the signal
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
T  
ShCoeffs<T>::reconstruct( const PointS2<S>& p) const {

  // Values of the SH basis functions at p
  sh_evaluator->shEvalBands( max_band, p.theta(), p.phi(), sh_fvals);

  T fval = 0.0;
  for( unsigned int band = 0; band < max_band; ++band) {
      fval += dot( coeffs[band], sh_fvals[band]);
  }

  return fval;
}

////////////////////////////////////////////////////////////////////////////////
//! Reconstruct the signal at p
//! @param  point where to reconstruct the signal
////////////////////////////////////////////////////////////////////////////////
template<class T>
template<class S>
T  
ShCoeffs<T>::reconstructYGradY( const PointS2<S>& p) const {

  // Values of the SH basis functions at p
  sh_evaluator->shEvalBands( max_band, p.theta(), p.phi(), sh_fvals, sh_grads);

  T fval = 0.0;
  T grad = std::pair<T,T>( 0.0, 0.0);
  for( unsigned int band = 0; band < max_band; ++band) {
    fval += dot( coeffs[band], sh_fvals[band]);
    grad += dot( coeffs[band], sh_grads);
  }

  return fval;
}

////////////////////////////////////////////////////////////////////////////////
//! Reconstruct the part of the signal in a specific band and its gradient 
//! at p where the point is those used with computeShVals() previously
//! @note Precondition: the SH basis function coefficients are stored.
//! @param  band  band for which to reconstruct the signal
//! @param  f_band  reconstructed function value for band 
//! @param  grad_band  reconstructed gradient for band
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShCoeffs<T>::reconstructYGradYBandAtP( const unsigned int band, T& f_band, 
                                       ShGrad& grad_band) const
{
  SH_COEFFS_DEBUG_ASSERT( band <= max_band, "reconstructYGradYBandAtP()", 
                                                "Invalid band.");

  f_band = dot( coeffs[band], sh_fvals[band]);
  grad_band = dot( coeffs[band], sh_grads[band]);
}

////////////////////////////////////////////////////////////////////////////////
//! Create the basis function coefficients corresponding to an acoustic source
//! @param  loc  location of the acoustic source
////////////////////////////////////////////////////////////////////////////////
template<class T>  
template<class S>
void  
ShCoeffs<T>::createAcousticSource( const unsigned int curr_max_band,
                                    const PointS2<S>& loc) 
{
  // normalization constant
  T normalization = sqrt(PI/(2.0*((T) curr_max_band)));

  // reconstruct function at source location
  sh_evaluator->shEvalBands( curr_max_band, loc.theta(), loc.phi(), coeffs);

  // multiply by bessel function
  for( unsigned int band = 0; band <= curr_max_band; ++band) {


    T bessel = gsl_sf_bessel_Jnu( ((T) band) + 0.5, curr_max_band);
    for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
      coeffs[band][i] *= bessel * normalization;
    }


    // calculate L2 norm of coefficients in this band
    T n = 0.0;
    for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
      n += coeffs[band][i] * coeffs[band][i];
    }
    n = sqrt( n);

    // normalize coefficients in this band
    // This is to ensure energy in each band is similar, in order
    // to allow a fair comparison across bands.
    for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
      coeffs[band][i] /= n;
    }

  }
}

////////////////////////////////////////////////////////////////////////////////
//! Create the basis function coefficients corresponding to an acoustic source
//! @param   max_sh_band   maximum band which should be considered for the 
//!                        source
////////////////////////////////////////////////////////////////////////////////
template<class T>  
template<class S>
void  
ShCoeffs<T>::createAcousticSource( const PointS2<S>& loc) {
  createAcousticSource( max_band, loc);
}

////////////////////////////////////////////////////////////////////////////////
//! l_p norm of the coefficient vector
//! By default, norm is calculated for all coefficients in all bands
//! (when band=-1)
//! If 'band' is passed, only calculates error for that specific band
//! @note By default the l_2 norm is computed
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShCoeffs<T>::norm( const unsigned int l_p, const int forband = -1) const {
  
  unsigned int start_index =0;
  unsigned int end_index = coeffs.size();
  
  // if band was passed, then calculate start/end indices
  // into coefficient vector to calculate error only for
  // this band
  if (forband >= 0) {
    start_index = forband;
    end_index = forband+1;
  }

  T n = 0.0;

  if( 1 == l_p) {
    for( unsigned int band = start_index; band < end_index; ++band) {
      for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
        n += std::abs(coeffs[band][i]);
      }
    } 
  }
  else if( 2 == l_p) {
    for( unsigned int band = start_index; band < end_index; ++band) {
      for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
        n += coeffs[band][i] * coeffs[band][i];
      }
    }
    n = sqrt( n);
  } 
  else if( std::numeric_limits<unsigned int>::infinity() == l_p) {
    for( unsigned int band = start_index; band < end_index; ++band) {
      for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
        n = max( n, coeffs[band][i]);
      }
    }
  }
  else {
    SH_COEFFS_DEBUG_ASSERT( false, "norm()", "Norm not implemented.");
  }

  return n;
}

////////////////////////////////////////////////////////////////////////////////
//! Get a specific coefficient
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T&
ShCoeffs<T>::operator()( const unsigned int band, const unsigned int i) const {
  
  SH_COEFFS_DEBUG_ASSERT( band < coeffs.size(), "operator()", "Invalid band.");
  SH_COEFFS_DEBUG_ASSERT( i < coeffs[band].size(), "operator()", 
                                                   "Invalid band index.");
  
  return coeffs[band][i];
}

////////////////////////////////////////////////////////////////////////////////
//! Set a specific coefficient
////////////////////////////////////////////////////////////////////////////////
template<class T>
T&  
ShCoeffs<T>::operator()( const unsigned int band, const unsigned int i) {

  SH_COEFFS_DEBUG_ASSERT( band < coeffs.size(), "operator()", "Invalid band.");
  SH_COEFFS_DEBUG_ASSERT( i < coeffs[band].size(), "operator()", 
                                                   "Invalid band index.");

  return coeffs[band][i];
}

////////////////////////////////////////////////////////////////////////////////
//! Get a pointer to the coefficients at a specific band.
////////////////////////////////////////////////////////////////////////////////
template<class T>
T* 
ShCoeffs<T>::getAddress( const unsigned int band) {

  SH_COEFFS_DEBUG_ASSERT(band < coeffs.size(), "getAddress()","Invalid band.");

  return &(coeffs[band].front());
}

////////////////////////////////////////////////////////////////////////////////
//! Get a pointer to the coefficients at a specific band.
////////////////////////////////////////////////////////////////////////////////
template<class T>
const T* 
ShCoeffs<T>::getAddress( const unsigned int band) const {

  SH_COEFFS_DEBUG_ASSERT(band < coeffs.size(), "getAddress()","Invalid band.");

  return &(coeffs[band].front());
}

////////////////////////////////////////////////////////////////////////////////
//! Get the maximum band for which the coefficient vector is initialized
////////////////////////////////////////////////////////////////////////////////
template<class T>
unsigned int 
ShCoeffs<T>::getMaxBand() const {
  return max_band;
}

////////////////////////////////////////////////////////////////////////////////
//! Set all elements to a specific value
////////////////////////////////////////////////////////////////////////////////
template<class T>
void  
ShCoeffs<T>::set( const T& val) {
  
  for( unsigned int band = 0; band < coeffs.size(); ++band) {
    for( unsigned int i = 0; i < coeffs[band].size(); ++i) {
      coeffs[band][i] = val;
    }
  }
}

