/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Core Spherical Harmonics functionality
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <cmath>
#include <cassert>

// includes, project
#include <config.h>
#include <legendre.h>

// namespace, unnamed
namespace {

} // end namespace, unnamed

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShEvaluator<T>::ShEvaluator() :
  max_band_init( 0),
  sin_table(),
  cos_table(),
  p_l(),
  dp_l()
{ }

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShEvaluator<T>::~ShEvaluator() { } 

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShEvaluator<T>*
ShEvaluator<T>::instance() {

  static ShEvaluator<T> inst;
  return &inst;
}

////////////////////////////////////////////////////////////////////////////////
//! Initialize
////////////////////////////////////////////////////////////////////////////////
template<class T>
void  
ShEvaluator<T>::init( const unsigned int max_band) {
  
  // nothing to do if already correctly initialized
  if( max_band == max_band_init) {
    return;
  }

  max_band_init = max_band;

  // temporaries for longitudal parts of SH basis functions
  sin_table.resize( max_band+1);
  cos_table.resize( max_band+1);

  // we have to keep track of the values of the Legendre polynomials for the 
  // the last two levels. Allocate the vectors once and for all since the 
  // maximum size is known
  p_l.resize( max_band+1);
  dp_l.resize( max_band+1);
}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate all Spherical Harmonics basis function in the specified band
//! @param band     Spherical Harmonics band.
//! @param theta    Spherical coordinate (latitude)
//! @param phi      Spherical coordinate (longitude)
//! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
//!                 linear indexing is from -l to +l)
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShEvaluator<T>::shEvalBand( const int& band, const T& theta, const T& phi, 
                            VecT& sh_vals) 
{
  DEBUG_ASSERT( (band <= (int)max_band_init), "ShEvaluator", "shEvalBand()",
                "Requested band larger than initialization.");

  // cache sin / cosine value, through the recursion these only have to be 
  // computed once
  T costheta = cos(theta);
  T sintheta = sin(theta);
  T cosphi = cos(phi);
  T sinphi = sin(phi);

  // fill precomputed longitudal components
  sin_table[0] = 0.0;
  cos_table[0] = 1.0;
  for( int i = 1; i <= band; ++i) {
    sin_table[i] = sinphi * cos_table[i-1] + cosphi * sin_table[i-1];
    cos_table[i] = cosphi * cos_table[i-1] - sinphi * sin_table[i-1];
  }

  // compute carry over using recursion
  T carry_over = Legendre::legendreComputeCarryOver<T>( band);

  // evaluate the Legendre polynomial for the new band using a recurrence
  // relationship
  Legendre::legendre( band, costheta, sintheta, p_l, carry_over);

  // normalization for current band   
  T normalization = sqrt(((T) (2*band+1) / (4*PI)));

  unsigned int m_index = 0;

  // sin bands (negative m)
  for( int m = band; m > 0; --m) {
    sh_vals[m_index] = normalization * p_l[m] * sin_table[m];
    ++m_index;
  }

  // zonal band
  sh_vals[m_index] = normalization * p_l[0];
  ++m_index;

  // cos bands (positive m)
  for( int m = 1; m <= band; ++m) {
    sh_vals[m_index] = normalization * p_l[m] * cos_table[m];
    ++m_index;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate all Spherical Harmonics basis functions between band_l and band_u
//! @param l      Spherical Harmonics band.
//! @param theta  Spherical coordinate (latitude)
//! @param phi    Spherical coordinate (longitude)
//! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
//!                 linear indexing is band after band and from -l to +l within
//!                 each band)
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShEvaluator<T>::shEvalBands( const int& band_l, const int& band_u,
                             const T& theta, const T& phi, 
                             ShCoeffsVecT& sh_vals) 
{
  DEBUG_ASSERT( (band_l <= (int)max_band_init), "ShEvaluator", "shEvalBands()",
                "Requested band larger than initialization.");
  DEBUG_ASSERT( (band_u <= (int)max_band_init), "ShEvaluator", "shEvalBands()",
                "Requested band larger than initialization.");
  DEBUG_ASSERT( (band_l <= band_u), "ShEvaluator", "shEvalBands()",
                "Bands have to be monotonic.");

  // use the specialized function if possible
  if( 0 == band_l) {
    shEvalBands( band_u, theta, phi, sh_vals);
    return;
  }

  // cache sin / cosine value, through the recursion these only have to be 
  // computed once
  T costheta = cos(theta);
  T sintheta = sin(theta);
  T cosphi = cos(phi);
  T sinphi = sin(phi);

  // fill precomputed longitudal components
  sin_table[0] = 0.0;
  cos_table[0] = 1.0;
  for( int i = 1; i <= band_u; ++i) {
    sin_table[i] = sinphi * cos_table[i-1] + cosphi * sin_table[i-1];
    cos_table[i] = cosphi * cos_table[i-1] - sinphi * sin_table[i-1];
  }

  // compute starting point for recursion
  T seed = Legendre::legendreComputeCarryOver<T>( band_l);

  // remaining bands
  for( int band = band_l; band <= band_u; ++band) {
    
    // evaluate the Legendre polynomial for the new band using a recurrence
    // relationship
    seed = Legendre::legendre( band, costheta, sintheta, p_l, seed);
 
    // normalization for current band   
    T normalization = sqrt(((T) (2*band+1) / (4*PI)));

    unsigned int m_index = 0;

    // sin bands (negative m)
    for( int m = band; m > 0; --m) {
      sh_vals[band][m_index] = normalization * p_l[m] * sin_table[m];
      ++m_index;
    }
    
    // zonal band
    sh_vals[band][m_index] = normalization * p_l[0];
    ++m_index;
    
    // cos bands (positive m)
    for( int m = 1; m <= band; ++m) {
      sh_vals[band][m_index] = normalization * p_l[m] * cos_table[m];
      ++m_index;
    }
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate all Spherical Harmonics basis functions up to and including l
//! @param l      Spherical Harmonics band.
//! @param theta  Spherical coordinate (latitude)
//! @param phi    Spherical coordinate (longitude)
//! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
//!                 linear indexing is band after band and from -l to +l within
//!                 each band)
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShEvaluator<T>::shEvalBands( const int& l, const T& theta, const T& phi, 
                             ShCoeffsVecT& sh_vals) 
{
  DEBUG_ASSERT( (l <= (int)max_band_init), "ShEvaluator", "shEvalBands()",
                "Requested band larger than initialization.");

  // cache sin / cosine value, through the recursion these only have to be 
  // computed once
  T costheta = cos(theta);
  T sintheta = sin(theta);
  T cosphi = cos(phi);
  T sinphi = sin(phi);

  // fill precomputed longitudal components
  sin_table[0] = 0.0;
  cos_table[0] = 1.0;
  for( int i = 1; i <= l; ++i) {
    sin_table[i] = sinphi * cos_table[i-1] + cosphi * sin_table[i-1];
    cos_table[i] = cosphi * cos_table[i-1] - sinphi * sin_table[i-1];
  }

  // compute Spherical Harmonics values up to band L using recurrence
  
  // band 0
  // the basis function is fixed so we can immediately store the value
  sh_vals[0][0] = std::sqrt(1.0/(4.0*PI));
  // early out if possible
  if( 0 == l) {
    return;
  }
  p_l[0] = 1;

  // carry forward for Legendre polynomial evaluation
  T seed = 1.0;

  // band = 1
  T normalization = sqrt(((T) (2*1+1) / (4*PI)));
  seed = Legendre::legendre( 1, costheta, sintheta, p_l, seed);
  sh_vals[1][0] = normalization * p_l[1] * sin_table[1];
  sh_vals[1][1] = normalization * p_l[0];
  sh_vals[1][2] = normalization * p_l[1] * cos_table[1];

  // remaining bands
  for( int band = 2; band <= l; ++band) {
    
    // evaluate the Legendre polynomial for the new band using a recurrence
    // relationship
    seed = Legendre::legendre( band, costheta, sintheta, p_l, seed);
 
    // normalization for current band   
    T normalization = sqrt(((T) (2*band+1) / (4*PI)));

    unsigned int m_index = 0;
    // sin bands (negative m)
    for( int m = band; m > 0; --m) {
      sh_vals[band][m_index] = normalization * p_l[m] * sin_table[m];
      ++m_index;
    }
    // zonal band
    sh_vals[band][m_index] = normalization * p_l[0];
    ++m_index;
    // cos bands (positive m)
    for( int m = 1; m <= band; ++m) {
      sh_vals[band][m_index] = normalization * p_l[m] * cos_table[m];
      ++m_index;
    }
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate all Spherical Harmonics basis functions up to and including l and 
//! their gradient
//! @param l      Spherical Harmonics band.
//! @param theta  Spherical coordinate (latitude)
//! @param phi    Spherical coordinate (longitude)
//! @param sh_vals  Values of the Spherical Harmonics basis functions (the 
//!                 linear indexing is band after band and from -l to +l within
//!                 each band)
//! @param sh_vals  Gradient (w.r.t. theta and phi) of sh_vals
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShEvaluator<T>::shEvalBands( const int& l, const T& theta, const T& phi, 
                             ShCoeffsVecT& sh_vals, ShGradVecT& d_sh_vals ) 
{
  DEBUG_ASSERT( (l <= (int)max_band_init), "ShEvaluator", "shEvalBands()",
                "Requested band larger than initialization.");

  // cache sin / cosine value, through the recursion these only have to be 
  // computed once
  T costheta = cos(theta);
  T sintheta = sin(theta);
  T cosphi = cos(phi);
  T sinphi = sin(phi);

  // fill precomputed longitudal components
  sin_table[0] = 0.0;
  cos_table[0] = 1.0;
  for( int i = 1; i <= l; ++i) {
    sin_table[i] = sinphi * cos_table[i-1] + cosphi * sin_table[i-1];
    cos_table[i] = cosphi * cos_table[i-1] - sinphi * sin_table[i-1];
  }

  // compute Spherical Harmonics values up to band L using recurrence
  
  // band 0
  // the basis function is fixed so we can immediately store the value
  sh_vals[0][0] = std::sqrt(1.0/(4.0*PI));
  d_sh_vals[0][0].first = 0.0;
  d_sh_vals[0][0].second = 0.0;

  // early out if possible
  if( 0 == l) {
    return;
  }
  p_l[0] = 1.0;
  dp_l[0] = 0.0;

  // carry forward for evaluation of Legendre polynomials
  T seed = 1.0;

  std::vector<T> work( 2*l+1);

  // band = 1
  T normalization = sqrt(((T) (2*1+1) / (4*PI)));
  seed = Legendre::legendre( 1, costheta, sintheta, p_l, dp_l, seed);
 
  sh_vals[1][0] = normalization * p_l[1] * sin_table[1];
  sh_vals[1][1] = normalization * p_l[0];
  sh_vals[1][2] = normalization * p_l[1] * cos_table[1];
  d_sh_vals[1][0].first  = normalization * dp_l[1] * sin_table[1];
  d_sh_vals[1][0].second = normalization * p_l[1] * cos_table[1];
  d_sh_vals[1][1].first  = normalization * dp_l[0]; 
  d_sh_vals[1][1].second = 0.0; 
  d_sh_vals[1][2].first  = normalization * dp_l[1] * cos_table[1];
  d_sh_vals[1][2].second = -1.0 * normalization * p_l[1] * sin_table[1];

  // remaining bands
  for( int band = 2; band <= l; ++band) {
    
    // evaluate the Legendre polynomial for the new band using a recurrence
    // relationship
    seed = Legendre::legendre( band, costheta, sintheta, p_l, dp_l, seed);
 
    // normalization for current band   
    T normalization = sqrt(((T) (2*band+1) / (4*PI)));

    unsigned int m_index = 0;
    // sin bands (negative m)
    for( int m = band; m > 0; --m) {
      sh_vals[band][m_index] = normalization * p_l[m] * sin_table[m];
      d_sh_vals[band][m_index].first = normalization * dp_l[m] * sin_table[m];
      d_sh_vals[band][m_index].second = m* normalization * p_l[m] * cos_table[m];
      ++m_index;
    }
    
    // zonal band
    sh_vals[band][m_index] = normalization * p_l[0];
    d_sh_vals[band][m_index].first = normalization * dp_l[0];
    d_sh_vals[band][m_index].second = 0.0;
    ++m_index;

    // cos bands (positive m)
    for( int m = 1; m <= band; ++m) {
      sh_vals[band][m_index] = normalization * p_l[m] * cos_table[m];
      d_sh_vals[band][m_index].first = normalization * dp_l[m] * cos_table[m];
      d_sh_vals[band][m_index].second = -m*normalization * p_l[m]* sin_table[m];
      ++m_index;
    }
  }

}





