/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Evaluation of Spherical Harmonics rotation
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <cstdlib>
#include <iomanip>
#include <fstream>
#include <ctime>
#include <cstdlib>
#include <cstdio>

// includes, boost
#include <boost/lexical_cast.hpp>

// includes, project
// #include <sh_rotation_rkhs.h>
// #include <sh_rotation_pinchon.h>
//#include <sh_rotation_pinchon_2.h>
#include <sh_rotation_rkhs.h>
#include <sh_rotation_rkhs_zonal.h>
#include <sh_rotation_pinchon.h>
#include <sh_rotation_pinchon_2.h>
#include <sh_rotation_pinchon_3.h>
#include <sh_rotation_spectral.h>
#include <sh_rotation_rkhs_spectral.h>
#include <sh_rotation_srkhs.h>
#include <sh_rotation_ivanic.h>
#include <sh_rotation_blanco.h>
#include <rotation.h>

// namespace, unnamed
namespace {

  //////////////////////////////////////////////////////////////////////////////
  //! Helper function for writing the different results to disk in Matlab format
  //////////////////////////////////////////////////////////////////////////////
  template<class FileType, class Array2D>
  void
  writeMatlabArray2D( FileType& file, 
                      const std::string& name, 
                      const Array2D& array) 
  {
    // execution times
    file << name << " = [";
    for( unsigned int k = 0; k < array.size(); ++k) {
      for( unsigned int i = 0; i < array[k].size(); ++i) {
        file << array[k][i] << ", ";
      }
      file << ";\n";
    }
    file << "];\n";
  }

  //////////////////////////////////////////////////////////////////////////////
  //! Create Matlab file name with current date and time
  //////////////////////////////////////////////////////////////////////////////
  void
  createFilenameMatlab( std::string& fname) {
  
    time_t now = time( NULL);
    std::tm* time = std::localtime( &now);

    fname += "results_";
    fname += boost::lexical_cast<std::string>(2000+time->tm_year-100);
    fname += "_";
    fname += boost::lexical_cast<std::string>(time->tm_mon+1);
    fname += "_";
    fname += boost::lexical_cast<std::string>(time->tm_mday);
    fname += "_";
    fname += boost::lexical_cast<std::string>(time->tm_hour);
    fname += "_";
    fname += boost::lexical_cast<std::string>(time->tm_min);
    fname += "_";
    fname += boost::lexical_cast<std::string>(time->tm_sec);
    fname += ".m";
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template< class T >
ShRotationBenchmark<T>::ShRotationBenchmark() :
  rotators(),
  exec_times(),
  errs_avg_l1(),
  errs_avg_l2(),
  errs_avg_linfty(),
  errs_max_l1(),
  errs_max_l2(),
  errs_max_linfty(),
  rotations_accuracy(),
  errs_accuracy()
{
  // generic points
  addRotator( new ShRotationRKHS<T>( 100));
  addRotator( new ShRotationRKHSZonal<T>( 100));
  addRotator( new ShRotationRKHS<T>( 100, "./data/optimized/", "optimized"));
  addRotator( new ShRotationRKHS<T>( 110, "./data/optimized/", "optimized"));
  addRotator( new ShRotationRKHS<T>( 200, "./data/optimized/", "optimized"));
  
  // equi-latitude
  addRotator( new ShRotationSRKHS<T>());

  // algorithms from the literature
  // addRotator( new ShRotationPinchon<T>());
  addRotator( new ShRotationPinchon2<T>());
  addRotator( new ShRotationPinchon3<T>());
  addRotator( new ShRotationIvanic<T>());
  addRotator( new ShRotationBlanco<T>());
  addRotator( new ShRotationSpectral<T>());

  // initialize random number generator
  srand( time( NULL));
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template< class T >
ShRotationBenchmark<T>::~ShRotationBenchmark() { 

  for( unsigned int i = 0; i < rotators.size(); ++i) {
    delete rotators[i];
    rotators[i] = NULL;
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Add a new rotator
////////////////////////////////////////////////////////////////////////////////
template< class T >
template<class S>
void
ShRotationBenchmark<T>::addRotator( S rotator) {

  rotators.push_back( rotator);
  exec_times.push_back( std::vector<T>());
  errs_avg_l1.push_back( std::vector<T>());
  errs_avg_l2.push_back( std::vector<T>());
  errs_avg_linfty.push_back( std::vector<T>());
  errs_max_l1.push_back( std::vector<T>());
  errs_max_l2.push_back( std::vector<T>());
  errs_max_linfty.push_back( std::vector<T>());
}

////////////////////////////////////////////////////////////////////////////////
//! Run experiment. Timing experiments are run for signals with 1 up to
//! max_band bands. Accuracy experiments are performed for a signal with
//! max_band bands and the error is reported per band.
//! @param  max_band  Maximum band of signals.
//! @param  num_rand_rotations   Number of random rotations in experiment
//! @param  num_iterations_timing  Number of iterations for timing measurements
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShRotationBenchmark<T>::runExperiment( const unsigned int max_band,
                                       const unsigned int num_rand_rotations,
                                       const unsigned int num_iterations_timing)
{
  // initialize the rotators
  for( unsigned int i = 0; i < rotators.size(); ++i) {
    rotators[i]->preload( max_band);
  }

  // Euler angles for rotation
  T t1, t2, t3 = 0.0;

  for( unsigned int band = 1; band <= max_band; ++band) {

    // coefficients
    ShCoeffs<T>  sh_coeffs( band);
    ShCoeffs<T>  sh_coeffs_ref( band);
    ShCoeffs<T>  sh_coeffs_rotated( band);

    // create random point (with uniform distribution) on S^2 as location for
    // the acoustic source
    T theta = PI/2 - getRandNormal<T>();
    T phi = getRandNormal<T>() * 2.0 * PI;
    PointS2<T> loc_source( theta, phi);

    // create the SH representation of the acoustic source
    sh_coeffs.createAcousticSource( band, loc_source);
    // sh_coeffs.set( 1.0);

    // temporary storage for the times
    std::vector<T>  times( rotators.size(), 0.0);
    std::vector<T>  errors_avg_l1( rotators.size(), 0.0);
    std::vector<T>  errors_avg_l2( rotators.size(), 0.0);
    std::vector<T>  errors_avg_li( rotators.size(), 0.0);
    std::vector<T>  errors_max_l1( rotators.size(), 0.0);
    std::vector<T>  errors_max_l2( rotators.size(), 0.0);
    std::vector<T>  errors_max_li( rotators.size(), 0.0);

    // do for all random rotations
    for( unsigned int k = 0; k < num_rand_rotations; ++k) {

      // generate random rotation
      rotation_generator.get( t1, t2, t3);

      // compute reference solution
      // rotate point
      PointS2<T> loc_source_rot( loc_source);
      loc_source_rot.rotate( t1, t2, t3);

      // create the SH representation of the acoustic source
      sh_coeffs_ref.createAcousticSource( band, loc_source_rot);

      // perform rotation with all rotators
      for( unsigned int i = 0; i < rotators.size(); ++i) {

        // clear to avoid bias due to previous iteration
        sh_coeffs_rotated.set( 0.0);

        T time = rotators[i]->rotatePreloaded( sh_coeffs, sh_coeffs_rotated,
                                               band, t1, t2, t3,
                                               num_iterations_timing);

        times[i] += time;

        // compute error

        // compute delta coefficient vector
        ShCoeffs<T> delta = sh_coeffs_ref - sh_coeffs_rotated;

        // Average of errors
        errors_avg_l1[i] += delta.norm( 1);
        errors_avg_l2[i] += delta.norm( 2);
        errors_avg_li[i] += delta.norm( ShCoeffs<T>::infty());
        
        // L_infty of errors
        errors_max_l1[i] = max( errors_max_l1[i], delta.norm( 1));
        errors_max_l2[i] = max( errors_max_l2[i], delta.norm( 2));
        errors_max_li[i] = max( errors_max_li[i], 
                                delta.norm( ShCoeffs<T>::infty()));

      }  // end for all rotators

    } // end for the given number of random rotations

    // normalize times by number of random rotations
    for( unsigned int i = 0; i < rotators.size(); ++i) {

      exec_times[i].push_back( times[i] / ((T) num_rand_rotations));
      // average error
      errs_avg_l1[i].push_back( errors_avg_l1[i] / ((T) num_rand_rotations));
      errs_avg_l2[i].push_back( errors_avg_l2[i] / ((T) num_rand_rotations));
      errs_avg_linfty[i].push_back(errors_avg_li[i]/((T)num_rand_rotations));
      // maximum error
      errs_max_l1[i].push_back( errors_max_l1[i]);
      errs_max_l2[i].push_back( errors_max_l2[i]);
      errs_max_linfty[i].push_back( errors_max_li[i]);
    }

    std::cout << "Finished band " << band << "." << std::endl;
  }  // end for all bands
}

////////////////////////////////////////////////////////////////////////////////
//!  Run experiment. Timing experiments are run for signals with 1 up to
//! max_band bands. Accuracy experiments are performed for a signal with
//! max_band bands and the error is reported per band.
//! @param  max_band  Maximum band of signals.
//! @param  num_rand_rotations   Number of random rotations in experiment
//! @param  num_iterations_timing  Number of iterations for timing measurements
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShRotationBenchmark<T>::
runExperimentAccuracy( const unsigned int max_band,
                       const unsigned int num_rand_rotations,
                       const unsigned int offset)
{
  // store rotations
  rotations_accuracy.resize( num_rand_rotations, std::vector<T>(3));
  // store complete errors for each rotation
  errs_accuracy.resize( rotators.size());

  // initialize
  for( unsigned int i = 0; i < rotators.size(); ++i) {

    // initialize rotators
    rotators[i]->preload( max_band);

    // initialize error storage for rotator
    errs_accuracy[i].resize( max_band+1);

    for( unsigned int j = 0; j <= max_band; ++j) {
      errs_accuracy[i][j].resize(3);
      for( unsigned int k = 0; k < 3; ++k) {
        errs_accuracy[i][j][k].resize( num_rand_rotations);
      }
    }
  }

  // Euler angles for rotation
  T t1,t2,t3 = 0.0;

  // coefficients
  ShCoeffs<T>  sh_coeffs( max_band);
  ShCoeffs<T>  sh_coeffs_ref( max_band);
  ShCoeffs<T>  sh_coeffs_rotated( max_band);

  // create random point (with uniform distribution) on S^2 as location for
  // the acoustic source
  T theta = PI/2 - getRandNormal<T>();
  T phi = getRandNormal<T>() * 2.0 * PI;
  PointS2<T> loc_source( theta, phi);

  // create the SH representation of the acoustic source
  sh_coeffs.createAcousticSource( max_band, loc_source);

  // skip a few rotations in the pseudo-random sequence
  for( unsigned int i=0; i<offset; i++) {
    rotation_generator.get(t1,t2,t3);
  }

  // do for all random rotations
  for( unsigned int k = 0; k < num_rand_rotations; ++k) {
    
    // generate random rotation
    rotation_generator.get( t1, t2, t3);
    rotations_accuracy[k][0] = t1;
    rotations_accuracy[k][1] = t2;
    rotations_accuracy[k][2] = t3;

    // compute reference solution
    // rotate point
    PointS2<T> loc_source_rot( loc_source);
    loc_source_rot.rotate( t1, t2, t3);

    // create the SH representation of the acoustic source
    sh_coeffs_ref.createAcousticSource( max_band, loc_source_rot);

    // perform rotation with all rotators
    for( unsigned int i = 0; i < rotators.size(); ++i) {

      // perform rotation
      rotators[i]->rotatePreloaded( sh_coeffs, sh_coeffs_rotated,
                                    max_band, t1, t2, t3, 1);
      // rotators[i]->rotate( sh_coeffs, sh_coeffs_rotated,
      //                     max_band, t1, t2 ,t3, 1);

      // compute delta coefficient vector
      ShCoeffs<T> delta = sh_coeffs_ref - sh_coeffs_rotated;

      // Average of errors for each individual band
      for (unsigned int band = 0; band <= max_band; band++) {
        errs_accuracy[i][band][0][k] = delta.norm(1,band);
        errs_accuracy[i][band][1][k] = delta.norm(2,band);
        errs_accuracy[i][band][2][k] = delta.norm(ShCoeffs<T>::infty(),band);
      }

    }  // end for all rotators
    
    std::cout << "Finished rotation " << k << "." << std::endl;

  } // end for the given numnber of random rotations

}

////////////////////////////////////////////////////////////////////////////////
//! Write the results to a Matlab file
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShRotationBenchmark<T>::
writeResultsMatlabAccuracy() const {

  // determine file name 
  std::string fname( "./results/accuracy_");
  createFilenameMatlab( fname);
  std::cout << "Writing to file " << fname << "\n";

  // open file
  std::ofstream file( fname.c_str());

  // write names (as cell array)
  file << "names = {";
  for( unsigned int i = 0; i < rotators.size(); ++i) {
    file << "'" << rotators[i]->name() << "', ";
  }
  file << "};\n\n";

  // write error data
  file << "errs = {";
  for( unsigned int i = 0; i < rotators.size(); ++i) {
    
    file << " {";
    for( unsigned int j = 0; j < 3; ++j) {

      file << " [";
      for( unsigned int band = 0; band < errs_accuracy[i].size(); ++band) {
        for( unsigned int k = 0; k < rotations_accuracy.size(); ++k) {
          file << errs_accuracy[i][band][j][k] << ", ";
        }
        file << ";\n";
      }
      file << "];\n";

    } // end norm
    file << "}, ";

  }
  file << "};\n\n";

  // write rotations
  file << "rotations = [";
  for( unsigned int i = 0; i < rotations_accuracy.size(); ++i) {
    file << rotations_accuracy[i][0] << ", " 
         << rotations_accuracy[i][1] << ", "
         << rotations_accuracy[i][2] << ";\n";
  }
  file << "];\n\n";

  // finish file IO
  file.close();

}

////////////////////////////////////////////////////////////////////////////////
//! Write the results to a Matlab file
////////////////////////////////////////////////////////////////////////////////
template< class T >
void
ShRotationBenchmark<T>::writeResultsMatlab() const {

  // determine file name 
  std::string fname( "./results/");
  createFilenameMatlab( fname);

  std::cout << "Results written to '" << fname << "'." << std::endl;

  // open file
  std::ofstream file( fname.c_str());

  // write names (as cell array)
  file << "names = {";
  for( unsigned int i = 0; i < rotators.size(); ++i) {
    file << "'" << rotators[i]->name() << "', ";
  }
  file << "};\n";

  // write execution times
  writeMatlabArray2D( file, "exec_times", exec_times);

  // write accuracy results
  writeMatlabArray2D( file, "err_avg_l1", errs_avg_l1);
  writeMatlabArray2D( file, "err_avg_l2", errs_avg_l2);
  writeMatlabArray2D( file, "err_avg_linfty", errs_avg_linfty);
  writeMatlabArray2D( file, "err_max_l1", errs_max_l1);
  writeMatlabArray2D( file, "err_max_l2", errs_max_l2);
  writeMatlabArray2D( file, "err_max_linfty", errs_max_linfty);
}

