
/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using the algorithm (and the
///               implementation of) Greengard and Gimbutas.
///
///  license    : See README for details. The rotation code used for this method
///               has been obtained from Leslie Greengard.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <cmath>
#include <vector>
#include <iostream>

// includes, boost
#include <boost/numeric/mtl/mtl.hpp>

// includes, project
#include <rotation.h>

// namespace, unnamed
namespace {

  // MTL matrix and vector convenience typedef's
  typedef  mtl::matrix::dense2D<FType>  Matrix;
  typedef  mtl::dense_vector<FType>  Vector;

  template<class M>
  void
  print_mat(M &D)
  {
    std::cout << std::setprecision(9);
    for (unsigned int i=0;i<num_rows(D);i++) {
      for (unsigned int j=0;j<num_cols(D);j++) {
        std::cout << D(i,j) << '\t';
      }
      std::cout << std::endl;
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  //! Calculate wigner-d matrix analytically for L=0 or L=1
  //////////////////////////////////////////////////////////////////////////////
  template<class M,class T>
  void
  wigner( const unsigned int& band, const T& beta, M &D)
  {
    T root2 = sqrt(2.0);
    if (band==0) {
      D(0,0)=1.0;
      return;
    } else if (band==1) {
      //analytic expression for 3x3 wigner-d matrix
      D(0,0) = cos(beta/2.0)*cos(beta/2.0);
      D(0,1) = 1.0/root2 * sin(beta);
      D(0,2) = sin(beta/2.0) * sin(beta/2.0);
      
      D(1,0) = - 1.0/root2 * sin(beta);
      D(1,1) = cos(beta);
      D(1,2) = 1.0/root2 * sin(beta);

      D(2,0) = sin(beta/2.0) * sin(beta/2.0);
      D(2,1) = - 1.0/root2 * sin(beta);
      D(2,2) = cos(beta/2.0) * cos(beta/2.0);

      return;
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  //! Calculate wigner d matrix for band L through recurrance on bands
  //  L-1 and L-2.  Numbered equations are from Blanco's paper
  //////////////////////////////////////////////////////////////////////////////
  template<class M,class T>
  void
  wigner( const unsigned int& band, const T& beta, M &D, M &dn1, M & dn2) {

    int L = (int)band;

    // for band < 2, we can calculate the matrix analytically
    if (L < 2) {
      wigner(L,beta,D);
      return;
    }

    // otherwise use recurrence scheme

    // precalculate some quantities
    T tanhalfbeta = tan(beta/2);
    T coshalfbeta = cos(beta/2);
    T cosbeta = cos(beta);

    for(  int m = 0; m <= (int)(L -2); m++) {
      for(  int mp = -m; mp <= m; mp++) {
        //eq 64
        T x = L * (2*L -1) / sqrt((T)(L*L - m*m)*(L*L-mp*mp));
        T t1 =(cosbeta - (m*mp)/(T)(L*(L-1))) * dn1(m+L-1,mp+L-1);
        T t2 = (sqrt((T)((L-1)*(L-1) - m*m) * ((L-1)*(L-1) - mp*mp))
                      / (T)((L-1) * (2*L-1))) * dn2(m+L-2,mp+L-2);
        x= x * (t1-t2);            
        D(m+L,mp+L) = x;
      }
    }

    //eq 65
    D(L+L,L+L) = pow(coshalfbeta,(2*L));
    
    // eq 66
    D(L-1+L,L-1+L) = (L * cosbeta - L +1) * pow(coshalfbeta,2*(L-1));

    //eq 67
    for (  int mp=L-1; mp >= -L; mp--) {
      D(L + L , mp + L) = - sqrt((T)(L+(mp+1))/(L-(mp+1)+1)) 
                                    * tanhalfbeta * D(L+L,L+mp+1);
    }

    //eq 68
    for (  int mp=L-2; mp >= 1-L; mp--) {

      T fac1 = -1.0 * tanhalfbeta 
	* (L*cosbeta-(mp+1)+1)/(L*cosbeta-(mp+1)) 
	* sqrt((T)(L+(mp+1))/(L-(mp+1)+1));

      D(L-1 + L, mp + L) = fac1 * D(L+L-1,L+mp+1);

      //special cases for singularities
      if (std::isinf(tanhalfbeta)) {
	if (mp ==1-L) {
	  D(L-1 + L, mp + L) = L-1;	  
	} else {
	  D(L-1 + L, mp + L) = 0;
	}
      } 

      if( std::isinf(fac1) || std::isnan(fac1)) {
	D(L-1 + L, mp + L) = 0;
      }

    }

    for( int m = 0; m <= L; m++) {
      for( int mp = 0; mp <= (m-1); mp++) {
        // calculate (-1)^(mp+m)
        T f = (((mp+m) % 2) ==0 ? 1 : -1);

        D(mp+L,m+L)=f*D(m+L,mp+L);
        D(mp+L,-m+L)= D(m+L,-mp+L);        
      }
    }

  }

  //////////////////////////////////////////////////////////////////////////////
  //! Calculate rotation matrix given wigner-d matrix and sin/cos lookup
  //  tables
  //////////////////////////////////////////////////////////////////////////////
  template<class M, class T>
  void
    getRotationMatrixBlanco(const unsigned int& band,
                            const std::vector<T>& sin_alpha,  
                            const std::vector<T>& cos_alpha, 
                            const std::vector<T>& sin_gamma, 
                            const std::vector<T>& cos_gamma, 
                            const M &D, M &rot_matrix)
  {
    T root2 = sqrt(2.0);
    int L = band;
    int m_fact = ((L % 2) == 0 ? -1  : 1);
    for (int m = -L; m <= L; m++) {
      // calculate (-1)^m
      m_fact  = -m_fact;
      for (int mp = -L; mp <= L; mp++) {
        T s1 = (mp >= 0 ? 1.0 : -1.0);
        T s2 = (m >= 0 ? 1.0 : -1.0);

        int absm=abs(m);
        int absmp=abs(mp);

        T d1 = D(absmp + L, absm + L);
        T d2 = m_fact * D(absm + L, -absmp + L);
  
        T omeg_m_alpha1 = 1.0;
        T omeg_m_alpha2 = 1.0;

        if (m < 0) {
          omeg_m_alpha1 = root2 * sin_alpha[absm];
          omeg_m_alpha2 = root2 * cos_alpha[absm];
        } else if (m > 0) {
          omeg_m_alpha1 = root2 * cos_alpha[absm];
          omeg_m_alpha2 = root2 * sin_alpha[absm];
        } else if (m==0) {
          omeg_m_alpha1 = 1;
          omeg_m_alpha2 = 1;
        }

        T omeg_mp_gamma1 = 1.0;
        T omeg_mp_gamma2 = 1.0;

        if (mp < 0) {
          omeg_mp_gamma1 = root2 * sin_gamma[absmp];
          omeg_mp_gamma2 = root2 * cos_gamma[absmp];
        } else if (mp > 0) {
          omeg_mp_gamma1 = root2 * cos_gamma[absmp];
          omeg_mp_gamma2 = root2 * sin_gamma[absmp];
        } else if (mp == 0) {
          omeg_mp_gamma1 = 1;
          omeg_mp_gamma2 = 1;
        }

        rot_matrix(mp+L, m+L) = 
           s1 * omeg_m_alpha1 * omeg_mp_gamma1 * (d1+d2)/2.0 
         - s2 * omeg_m_alpha2 * omeg_mp_gamma2 * (d1 - d2)/2.0;
      }
    }
  }

} // end namespace, unnamed


/*static*/ 
template<class T>
const std::string ShRotationBlanco<T>::class_name = "ShRotationBlanco";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationBlanco<T>::ShRotationBlanco() :
  ShRotationBase<T>(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationBlanco<T>::~ShRotationBlanco() { 

  cutDeleteTimer( timer);
}


////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationBlanco<T>::preload( const unsigned int /*max_band*/) { }

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationBlanco<T>::rotate( const ShCoeffs<T>& sh_coeffs, 
                             ShCoeffs<T>& sh_coeffs_rotated, 
                             const unsigned int max_band,
                             const T& t1, const T& t2, const T& t3,
                             const unsigned int num_iterations)
{
  // rename variables to correspond to equations in Blanco paper
  T alpha = -t1;
  T beta = -t2;
  T gamma = -t3;

  // helper variable 
  std::vector<T>  temp( 2*max_band+1, 0.0);

  // first band is invariant under rotation
  sh_coeffs_rotated(0,0) = sh_coeffs(0,0);
  // start timer
  cutStartTimer( timer);

  for (unsigned int it = 0; it < num_iterations; it++) {

    // Sin/Cos lookup tables
    std::vector<T> sin_alpha(max_band+1);
    std::vector<T> cos_alpha(max_band+1);
    std::vector<T> sin_gamma(max_band+1);
    std::vector<T> cos_gamma(max_band+1);
    sin_alpha[0] = 0;
    sin_gamma[0] = 0;
    cos_alpha[0] = 1;
    cos_gamma[0] = 1;

    T sin_alpha1 = sin(alpha);
    T cos_alpha1 = cos(alpha);
    T sin_gamma1 = sin(gamma);
    T cos_gamma1 = cos(gamma);

    // maintain pointers to sequence of wigner matrices
    Matrix *dn1=new Matrix(1,1);
    Matrix *dn2=new Matrix(3,3);
    Matrix *D=0;
    wigner(0,beta,*dn1);
    wigner(1,beta,*dn2);

    for( unsigned int L = 1; L <= max_band; ++L) {

      // calculate sin/cos lookup tables entries
      const unsigned int band_size = 2*L+1;
      D = new Matrix(band_size, band_size);

      sin_alpha[L] = sin_alpha1 * cos_alpha[L-1] + cos_alpha1 * sin_alpha[L-1];
      cos_alpha[L] = cos_alpha1 * cos_alpha[L-1] - sin_alpha1 * sin_alpha[L-1];

      sin_gamma[L] = sin_gamma1 * cos_gamma[L-1] + cos_gamma1 * sin_gamma[L-1];
      cos_gamma[L] = cos_gamma1 * cos_gamma[L-1] - sin_gamma1 * sin_gamma[L-1];

      if (L < 2) {
        wigner(L,beta,*D);
      } else {
        wigner(L,beta,*D,*dn1,*dn2);
      }

      Matrix cbm_band( band_size, band_size);
      getRotationMatrixBlanco( L,sin_alpha, cos_alpha, sin_gamma, cos_gamma, 
                               *D,cbm_band);

      // construct vector of SH coefficients in the current band
      // first create temporary
      memcpy( &(temp.front()), sh_coeffs.getAddress( L), sizeof(T)*band_size);
      Vector coeffs( band_size, &(temp.front()));

      // create vector for result
      Vector coeffs_rotated( 2*L+1, sh_coeffs_rotated.getAddress(L));

      // perform matrix vector multiplication
      mult( cbm_band, coeffs, coeffs_rotated);

      // throw away wigner matrix for band L-2
      if (dn2) { delete dn2; dn2=0; }

      // maintain history of past L-1 and L-2 wigner matrices for next iteration
      dn2 = dn1;
      dn1 = D;
    }
  
    // delete remaining wigner matrices
    if (dn1) { delete dn1; dn1=0; }
    if (dn2) { delete dn2; dn2=0; }
  
  } // end timing iterations

  cutStopTimer( timer);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) num_iterations);

  cutResetTimer( timer);

  return ret_val;
}


////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationBlanco<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                         ShCoeffs<T>& sh_coeffs_rotated, 
                                         const unsigned int max_band,
                                         const T& t1, const T& t2, const T& t3,
                                         const unsigned int num_iterations)
{
  //std::cout << "WARNING :: ShRotationBlanco<T>::rotatePreloaded() : "
  //          << "No precomputation implemented." << std::endl; 

  return rotate( sh_coeffs, sh_coeffs_rotated, max_band,
                 t1, t2, t3, num_iterations); 
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationBlanco<T>::name() const {
  return class_name;
}
  
