/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Spherical Harmonics rotation using RKHS
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iomanip>

// includes, libraries
#include <boost/lexical_cast.hpp>

// includes, project
#include <rotation.h>

// namespace, unnamed
namespace {

} // end namespace, unnamed

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
//! @param oversampling factor
//! @param  path   path to data files
//! @param  name_addendum  additional string for class name
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationRKHS<T>::ShRotationRKHS( const unsigned int oversampling_factor,
                                   const std::string path,
                                   const std::string  name_addendum) :
  ShRotationBase<T>(),
  os_factor( oversampling_factor),
  num_bands_preloaded( 0),
  set_sampling_nested( false),
  set_sampling(),
  cbm(),
  fvals(),
  path_data( path),
  timer( 0)
{
  class_name = std::string( "ShRotationRKHS (");
  class_name += boost::lexical_cast<std::string>(os_factor);
  class_name += std::string(")");
  if( 0 != name_addendum.length()) {
    class_name += ", ";
    class_name += name_addendum;
  }

  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationRKHS<T>::~ShRotationRKHS() {
  cutDeleteTimer(timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Load rotation matrices and associated sets of sampling
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'sh_rotation_rkhs_b*.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHS<T>::preload( const unsigned int max_band) {

  // set default file name if none is specified
  if( 0 == path_data.length()) {
    path_data = "data/";
  }
  std::string fname( path_data);
  fname += "sh_rotation_rkhs_b";

  // set band 0 (the elements are only place-holders, for rotation the basis
  // function coefficients for band 0 can simply be copied
  set_sampling.push_back( std::vector< PointS2<T> >());
  Matrix  cbm_c( 1, 1);
  cbm_c(0,0) = 1;
  cbm.push_back( cbm_c);

  // do for all bands greater than 0
  for( unsigned int band = 1; band <= max_band; ++band) {

    //calculate dimensions of matrix and sampling set vector
    unsigned int dim_band = 2*band+1;
    unsigned int num_samples = numSamplesOversampling( band, os_factor);

    Matrix  cbm_c(dim_band,num_samples);
    std::vector< PointS2<T> >  set_sampling_c(num_samples);

    loadCbm( band, fname, cbm_c, set_sampling_c);
    set_sampling.push_back( set_sampling_c);
    cbm.push_back( cbm_c);

  } // end reading input data

  num_bands_preloaded = max_band;

  // determine if the set of sampling is nested
  checkSetOfSamplingNested();

  // allocate memory for the temporaries
  prealloc( os_factor);
}

////////////////////////////////////////////////////////////////////////////////
//! Load rotation matrices and associated sets of sampling
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'sh_rotation_rkhs_b*.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHS<T>::loadCbm( const unsigned int band, std::string fname, 
                            Matrix& cbm_c,
                            std::vector< PointS2<T> >& set_sampling_c) 
{
  std::string fname_c( fname);
  fname_c += boost::lexical_cast<std::string>( band);
  fname_c += "_os" + boost::lexical_cast<std::string>( os_factor);
  fname_c += ".bin";

  std::ifstream fs( fname_c.c_str());
  // check if opening the file succeeded
  if( ! fs.good()) {
    std::cerr << "WARNING: Failed to open input file for band "
              << band << ": '" << fname_c << "'.\n";
    return;
  }

  // read the data

  unsigned int num_samples = numSamplesOversampling(band,os_factor);

  unsigned int dim_band = 2 * band + 1;

  T *memblock = new T[dim_band * num_samples];

  fs.read((char*)memblock,sizeof(T)* num_samples * 3);

  T *block = memblock;

  for( unsigned int i = 0; i < num_samples; ++i) {
    set_sampling_c[i].x() = *block++;
    set_sampling_c[i].y() = *block++;
    set_sampling_c[i].z() = *block++;
  }

  fs.read((char*)memblock,sizeof(T)* dim_band * num_samples);

  block = memblock;


  for( unsigned int i = 0; i < dim_band; ++i) {
    for( unsigned int j = 0; j < num_samples; ++j) {
      cbm_c(i,j) = *block++;
    }
  }

  fs.close();
  delete[] memblock;
}

////////////////////////////////////////////////////////////////////////////////
//! Check if the point set in \a set_sampling is nested and set
//! \a set_sampling_nested appropriately
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHS<T>::checkSetOfSamplingNested() {

  // Determine if the set of sampling is nested
  bool is_nested = true;
  for( unsigned int band = 1; band < num_bands_preloaded; ++band) {
    for( unsigned int i = 0; i < (2*band+1); ++i) {
      if( set_sampling[num_bands_preloaded-1][i] != set_sampling[band][i]) {
        is_nested = false;
        break;
      }
    }
    if( false == is_nested) {
      break;
    }
  }
  set_sampling_nested = is_nested;

#if DEBUG
  if( true == set_sampling_nested) {
    std::cout << "Set of sampling is nested.\n";
  }
  else {
    std::cout << "Set of sampling is NOT nested.\n";
  }
#endif

}

////////////////////////////////////////////////////////////////////////////////
//! Determine number of samples used in a band for a given oversampling rate
////////////////////////////////////////////////////////////////////////////////
template<class T>
unsigned int
ShRotationRKHS<T>::numSamplesOversampling( unsigned int band, 
                                           unsigned int os_factor) 
{
  double os_factor_t  = ((double) os_factor) / 100.0;
  return ceil((2*band+1) * os_factor_t);
}

////////////////////////////////////////////////////////////////////////////////
//! Allocate memory for \a fvals, the temporary storage for function values
//! during rotation
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHS<T>::prealloc( const unsigned int os_factor) {

  fvals.resize( num_bands_preloaded+1);
  for( unsigned int band = 0; band < fvals.size(); ++band) {
    fvals[band] = Vector( numSamplesOversampling( band, os_factor));
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time for function values (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3  Euler ZYZ rotation angles
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationRKHS<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs,
                                    ShCoeffs<T>& sh_coeffs_rotated,
                                    const unsigned int max_band,
                                    const T& t1, const T& t2, const T& t3,
                                    const unsigned int iters)
{
  DEBUG_ASSERT( ( max_band <= num_bands_preloaded), "ShRotationRKHS",
                  "rotatePreloaded()", "Not enough bands preloaded.");

  // compute rotation matrix
  Matrix R(3,3);
  // the rotation has to be applied inversely to the sample points
  Rotation::getRotationMatrixEulerZYZ( -t3, -t2, -t1, R);

  // internal coefficient vector for computations
  ShCoeffs<T>  sh_coeffs_r( sh_coeffs);

  // start timer
  cutStartTimer( timer);

  for( unsigned int i = 0; i < iters; ++i) {

    // reconstruct the function at the rotated set of sampling
    if( set_sampling_nested) {
      computeFValsNested( sh_coeffs_r, max_band, R);
    }
    else {
      computeFValsNotNested( sh_coeffs_r, max_band, R);
    }

    // apply the change of basis matrix back to Spherical Harmonics

    // the coefficient of band zero is invariant under rotation
    sh_coeffs_rotated(0,0) = sh_coeffs(0,0);

    // iterate over all bands and apply change of basis matrices
    for( unsigned int band = 1; band <= max_band; ++band) {

      // create temporary vector
      Vector res( 2*band+1, sh_coeffs_rotated.getAddress(band));

      // perform matrix vector multiplication
      mult( cbm[band], fvals[band], res);
    }
  }

  // stop timer
  cutStopTimer( timer);

  // validate rotation data
  // validateRotationMatlab(max_band, sh_coeffs, sh_coeffs_rotated, t1, t2, t3);

  // get timer value and then reset for next call
  T ret_val = cutGetTimerValue( timer) / ((T) iters);
  cutResetTimer( timer);

  return ret_val;
}


////////////////////////////////////////////////////////////////////////////////
//! Evaluate per band signals at rotated sample points for an unnested set 
//! of sampling.
//! @param  sh_coeffs_r  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  R          rotation matrix
//! @param  iters      Number of iterations (for computing rotation
//!                    matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHS<T>::
computeFValsNotNested( ShCoeffs<T>& sh_coeffs_r,
                       const unsigned int max_band, const Matrix& R) 
{
  // for all bands
  for( unsigned int band = 1; band <= max_band; ++band) {

    // for all sampling points in the current band
    PointS2<T>  p_r;
    for( unsigned int i = 0; i < set_sampling[band].size(); ++i) {

      // get current point and rotate it
      p_r = set_sampling[band][i];
      p_r.rotate( R);

      // reconstruct function at current point
      sh_coeffs_r.computeYBand( band, p_r);
      fvals[band][i] = sh_coeffs_r.reconstructBandAtP( band);
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate per band signals at rotated sample points for a nested set 
//! of sampling.
//! @param  sh_coeffs_r  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  R          rotation matrix
//! @param  iters      Number of iterations (for computing rotation
//!                    matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationRKHS<T>::
computeFValsNested( ShCoeffs<T>& sh_coeffs_r,
                    const unsigned int max_band, const Matrix& R) 
{
  PointS2<T>  p_r;
  for( unsigned int i = 0; i < set_sampling[max_band].size(); ++i) {

    // get current point and rotate it
    p_r = set_sampling[max_band][i];
    p_r.rotate( R);

    // reconstruct function at current point
    // min_band = (i-1) / 2 (from dim_band = 2*band+1) but C indexing changes 
    // this (conveniently) to i/2
    unsigned int min_band = 
      (unsigned int) floor((((T) i)/(os_factor/100.0)) / 2.0);
    sh_coeffs_r.computeY( min_band, max_band, p_r);

    for( unsigned int band = 1; band <= max_band; ++band) {

      // the dimension of the invariant subspace increases with the band and
      // hence also the number of vertices which have to be considered;
      // conversely, for lower bands we do not have to consider all vertices
      // only calculate fvals[band][i] for bands that make use of this sampling
      //point
      if (i < numSamplesOversampling( band, os_factor)) {
        // reconstruct the function value *for* the basis functions from
        // the current band
        DEBUG_ASSERT( band >= min_band, "ShRotationRKHS", "rotatePreloaded()",
                                        "'min_band' too small.");
        fvals[band][i] = sh_coeffs_r.reconstructBandAtP( band);
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time for function values (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3  Euler ZYZ rotation angles
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationRKHS<T>::rotate( const ShCoeffs<T>& sh_coeffs,
                           ShCoeffs<T>& sh_coeffs_rotated,
                           const unsigned int max_band,
                           const T& t1, const T& t2, const T& t3,
                           const unsigned int num_iterations)
{
  preload( max_band);

  return rotatePreloaded( sh_coeffs, sh_coeffs_rotated, max_band, t1, t2, t3, 
                          num_iterations);
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string&
ShRotationRKHS<T>::name() const {
  return class_name;
}
  
