/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2009      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : sh
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: 
///
///  license    : See README for details. 
///
////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <vector>
#include <iostream>

// includes, boost
#include <fftw3.h>

// includes, project
#include <point_s2.h>
#include <rotation.h>
#include <spectral_util.h>

using namespace Spectral;

// namespace, unnamed
namespace {

} // end namespace, unnamed

/*static*/ 
template<class T>
const std::string 
ShRotationSRKHS<T>::class_name = "ShRotationSRKHS";

////////////////////////////////////////////////////////////////////////////////
//! Constructor, default
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationSRKHS<T>::ShRotationSRKHS() :
  ShRotationBase<T>(),
  zeta( 9.0*PI/20.0),
  cos_zeta( std::cos(zeta)),
  sin_zeta( std::sin(zeta)),
  f_vals(),
  timer( 0)
{
  cutCreateTimer( &timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Destructor
////////////////////////////////////////////////////////////////////////////////
template<class T>
ShRotationSRKHS<T>::~ShRotationSRKHS() { 

  cutDeleteTimer( timer);
}

////////////////////////////////////////////////////////////////////////////////
//! Initialize rotation, in particular load precomputed data
//! @param  max_band  maximum band to pre-load
//! @param  fname  file name containing the data needed for rotation
//!                the default name is 'data/pinchonNumMatJ.dat'
////////////////////////////////////////////////////////////////////////////////
template<class T>
void 
ShRotationSRKHS<T>::preload( const unsigned int max_band ) {

  // allocate memory
  f_vals.resize( max_band+1);
  for( unsigned int band = 0; band <= max_band ; ++band) {
    f_vals[band] = std::vector<T>( 2*max_band+1);
  }
 
  // helper variable for evaluation of associated Legendre polynomials
  T p_nm_help = 1.0;

  p_nm_scaled.resize( max_band+1);

  for( unsigned int band = 0; band <= max_band; ++band) {

    // allocate memory
    p_nm_scaled[band].resize( 2*band+1);
    
    // compute Legendre polynomial values
    p_nm_help = Legendre::legendre( band, cos_zeta, sin_zeta, 
                                    p_nm_scaled[band], p_nm_help);

    // scale Legendre polynomial values
    T scale_p_nm = 0.5 * sqrt((2*band+1) / (4*PI));
    for( unsigned int i = 0; i < (2*band+1); ++i) {
      p_nm_scaled[band][i] *= scale_p_nm;
    }
    p_nm_scaled[band][0] *= 2;
  }

}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @note The implementation is similar to those by Greengard and Gimbutas.
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationSRKHS<T>::rotate( const ShCoeffs<T>& sh_coeffs_in, 
                             ShCoeffs<T>& sh_coeffs_r, 
                             const unsigned int max_band,
                             const T& t1, const T& t2, const T& t3,
                             const unsigned int num_iterations)
{
  // allocate temporaries

  // points on the equator
  std::vector< PointS2<T> > pts;
  const unsigned int num_samples = 2*max_band + 1;

  // allocate memory for discrete Fourier transform
  fftw_complex* f_vals_dft = 
    (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * num_samples);
  T* f_vals_in = (T*) fftw_malloc(sizeof(T) * num_samples);

  // determine optimized FFTW plans
  fftw_plan p1 = fftw_plan_dft_r2c_1d( num_samples, f_vals_in, f_vals_dft,
                                       FFTW_MEASURE);

  //! temporaries needed for Legendre polynomials and its derivatives which are 
  //! needed for recovering the Spherical Harmonics coefficients
  std::vector<T> p_nm( max_band+1);

  // create modifiable coefficient vector
  ShCoeffs<T> sh_coeffs( sh_coeffs_in);

  // construct *inverse* rotation matrix and rotate points
  Matrix R(3,3);
  Rotation::getRotationMatrixEulerZYZ( -t3, -t2, -t1, R);
    
  // start timer
  cutStartTimer( timer);

  for( unsigned int iters = 0; iters < num_iterations; ++iters) {

    // generate points on the equator and rotate the points
    createPointsEquiLatitudeR( num_samples, zeta, pts, R);

    // compute sample values
    computeSampleValues( sh_coeffs, max_band, pts);

    // the zero-th band is truly invariant under rotation
    sh_coeffs_r(0,0) = sh_coeffs(0,0);

    // recover all bands
    for( unsigned int band = 1; band <= max_band; ++band) {

      // compute Fourier transform for function values and projected gradient

      // copy input date to fft arrays
      memcpy( f_vals_in, &(f_vals[band].front()), sizeof(T)*num_samples);

      // execute fft
      fftw_execute(p1); 

      // Finally recover SH coefficients. The discrete Fourier transform yields
      // band+1 complex coefficients.

      for( unsigned int i = 0; i < (band+1); ++i) {

        // Recover Spherical Harmonics basis function coefficients from the 
        // result of the discrete Fourier transform
        sh_coeffs_r(band,band-i) = valSin( i, f_vals_dft, p_nm_scaled[band]);
        sh_coeffs_r(band,band-i) /= num_samples;
        sh_coeffs_r(band,band+i) = valCos( i, f_vals_dft, p_nm_scaled[band]);
        sh_coeffs_r(band,band+i) /= num_samples;
      }
    }

  } // end iterations for accurate timing

  // get timer value and then reset for next call
  cutStopTimer( timer);
  T time = cutGetTimerValue( timer) / ((T) num_iterations);
  cutResetTimer( timer);

  // cleanup memory
  fftw_destroy_plan( p1);
  fftw_free( f_vals_dft);
  fftw_free( f_vals_in);

  return time;
}

////////////////////////////////////////////////////////////////////////////////
//! Rotate a signal represented in Spherical Harmonics
//! @return  Computation time of rotation matrix (ms).
//! @param  sh_coeffs  Spherical Harmonics coefficient vector (unrotated)
//! @param  sh_coeffs_rotated  Spherical Harmonics coefficient vector 
//!                            (after rotation)
//! @param  max_band   Maximum band of \a sh_coeffs to be rotated (that is 
//!                    \a sh_coeffs has to contain at least max_band bands)
//! @param  t1 / t2 / t3   ZYZ Euler rotation angles.
//! @param  num_iterations  Number of iterations (for computing rotation
//!                         matrices and measuring execution time)
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
ShRotationSRKHS<T>::rotatePreloaded( const ShCoeffs<T>& sh_coeffs, 
                                         ShCoeffs<T>& sh_coeffs_rotated, 
                                         const unsigned int max_band,
                                         const T& t1, const T& t2, const T& t3,
                                         const unsigned int num_iterations)
{
  return rotate( sh_coeffs, sh_coeffs_rotated, max_band,
                 t1, t2, t3, num_iterations); 
}

////////////////////////////////////////////////////////////////////////////////
//! Compute sample values for the function and its derivative
//! @param max_band  maximum band
//! @param pts       sample points (rotated)
//! @param beta      rotation angle (a 
////////////////////////////////////////////////////////////////////////////////
template<class T>
void
ShRotationSRKHS<T>::
computeSampleValues( ShCoeffs<T>& sh_coeffs, const unsigned int max_band, 
                     std::vector< PointS2<T> >& pts)
{
  // for all sampling locations (the points on the equator with the *inverse*
  // rotation) applied
  for( unsigned int i = 0; i < pts.size(); ++i) {

    // compute values of Spherical Harmonics basis functions and gradient
    // at current point
    sh_coeffs.computeY( max_band, pts[i]);

    // the dimension of the invariant subspace increases with the band and
    // hence also the number of vertices which have to be considered;
    // conversely, for lower bands we do not have to consider all vertices
    for( unsigned int band = 1; band <= max_band; ++band) {
      // reconstruct the function value *for* the basis functions from
      // the current band
      f_vals[band][i] = sh_coeffs.reconstructBandAtP( band);
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//! Get the name of the rotation algorithm
////////////////////////////////////////////////////////////////////////////////
template<class T>
const std::string& 
ShRotationSRKHS<T>::name() const {
  return class_name;
}
  
