/// @file
////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2008      University of Toronto
///
////////////////////////////////////////////////////////////////////////////////
///
///  module     : core
///
///  author     : lessig@dgp.toronto.edu
///
///  project    : Spherical Harmonics rotation
///
///  description: Evaluate Legendre polynomials and their derivatives.
///
///  license    : See README for details.
///
////////////////////////////////////////////////////////////////////////////////

#ifndef _LEGENDRE_H_
#define _LEGENDRE_H_

// includes, system
#include <cmath>
#include <limits>

namespace Legendre {

////////////////////////////////////////////////////////////////////////////////
//! Compute carry over for legendre if a specific band is to be computed without
//! the previous bands being relevant
//! @return  carry over of interest
//! @param   degree for which Legendre polynomials are to be computed
////////////////////////////////////////////////////////////////////////////////
template<class T>
T
legendreComputeCarryOver( const unsigned int degree) {

  T f1 = 1.0;
  for( unsigned int i = 1; i < degree; ++i) {
    T i_f = (T) i;
    f1 = f1 * (2*i_f-1) / (2*i_f);
  }

  return f1;
}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate Legendre polynomials and its derivative at theta \in [-1,1].
//! @note The function is intended to be used repeatedly to compute all degress
//!       starting from degree zero to some final degree. The carry forward
//!       allows to exploit the recursive structure of Legendre polynomials
//!       without the burden of having to store all function values for all
//!       degrees. If only a specific band is of interest then the carry forward
//!       can be computed using legendreComputeCarryOver()
//! @note The derivatives are scaled by -sin(theta)
//! @return  carry over for next higher band
//! @param degree  degree for which to evaluate Legendre polynomials and its
//!                derivatives
//! @param cos_theta  cos(theta)
//! @param sin_theta  sin(theta)
//! @param p_nm  function values of Legendre polynomials (size has to be 
//!              (degree+1) or larger)
//! @param dp_nm  derivatives of Legendre polynomials (size has to be 
//!               (degree+1) or larger)
//! @param carry_forward  carry forward from degree-1 (if computations are stated
//!                       at degree one then the carry over should be initialized
//!                       to 1.0
////////////////////////////////////////////////////////////////////////////////
template<class ArrayT, class T>
T
legendre( const unsigned int degree, const T cos_theta, const T sin_theta,
           ArrayT& p_nm, ArrayT& dp_nm, T carry_forward ) 
{
  if( 0 == degree) {
    p_nm[0] = dp_nm[0] = 0.0;
    return 1.0;
  }  

  T degree_f = (T) degree;

  // handle positions close to the pole
  // for test purposes use for example 
  // theta (+/-) std::numeric_limits<T>::epsilon() / 2.0
  if( std::abs(sin_theta) < std::numeric_limits<float>::epsilon()) {
    
    // almost all entries are zero
    for( unsigned int i = 0; i <= degree; ++i) {
      p_nm[i] = dp_nm[i] = 0.0;
    }

    // only the sin_theta < 0.0 (south pole) and odd degree's yield -1.0
    T sign_factor = (sin_theta >= 0.0) ? 1.0 : (0==(degree+1) % 2) ? -1.0 : 1.0;
    p_nm[0] = sign_factor;
    dp_nm[1] = sqrt(degree_f*(degree_f+1.0))/2.0 * sqrt(2)* sign_factor;

    // carry over is irrelevant
    return 1.0;
  }

  // handle regular points

  carry_forward *= (2.0*degree_f-1.0) / (2.0*degree_f);
  p_nm[degree] = sqrt(carry_forward);
  dp_nm[degree] = 0.0;

  for( int m = (int) (degree-1); m >= 0; --m) {
    p_nm[m] = -(sin_theta * dp_nm[m+1] + 2 * (m+1) * cos_theta * p_nm[m+1]);

    T m_f = (T) m+1;
    T scale2 = sqrt((degree_f + m_f) * (degree - m_f + 1));
    p_nm[m] /= scale2;

    dp_nm[m] = sin_theta * p_nm[m+1] * scale2; 
  }

  T factor = (0 == (degree % 2)) ? 1.0 : -1.0;
  p_nm[0] *= factor;
  dp_nm[0] *= factor;
  factor *= -sqrt(2);
  T f3 = 1.0;
  for( unsigned int m = 1; m < degree+1; ++m) {
    dp_nm[m] = (sin_theta * dp_nm[m] + m * cos_theta * p_nm[m]) * f3 * factor;
    p_nm[m] *= sin_theta * f3 * factor; 
    f3 *= sin_theta;
    factor *= -1.0;
  }

  return carry_forward; 
}

////////////////////////////////////////////////////////////////////////////////
//! Evaluate Legendre polynomials at theta \in [-1,1].
//! @note The function is intended to be used repeatedly to compute all degress
//!       starting from degree zero to some final degree. The carry forward
//!       allows to exploit the recursive structure of Legendre polynomials
//!       without the burden of having to store all function values for all
//!       degrees. If only a specific band is of interest then the carry forward
//!       can be computed using legendreComputeCarryOver()
//! @return  carry over for next higher band
//! @param degree  degree for which to evaluate Legendre polynomials and its
//!                derivatives
//! @param cos_theta  cos(theta)
//! @param sin_theta  sin(theta)
//! @param p_nm  function values of Legendre polynomials (size has to be 
//!              (degree+1) or larger)
//! @param carry_forward  carry forward from degree-1 (if computations are stated
//!                       at degree one then the carry over should be initialized
//!                       to 1.0
////////////////////////////////////////////////////////////////////////////////
template<class ArrayT, class T>
T
legendre( const unsigned int degree, const T cos_theta, const T sin_theta,
           ArrayT& p_nm, T carry_forward ) 
{
  if( 0 == degree) {
    p_nm[0] = 0.0;
    return 1.0;
  }  

  T degree_f = (T) degree;

  // handle positions close to the pole
  // for test purposes use for example 
  // theta (+/-) std::numeric_limits<T>::epsilon() / 2.0
  if( std::abs(sin_theta) < std::numeric_limits<float>::epsilon()) {
    
    // almost all entries are zero
    for( unsigned int i = 0; i <= degree; ++i) {
      p_nm[i] = 0.0;
    }

    // only the sin_theta < 0.0 (south pole) and odd degree's yield -1.0
    T sign_factor = (sin_theta >= 0.0) ? 1.0 : (0==(degree+1) % 2) ? -1.0 : 1.0;
    p_nm[0] = sign_factor;

    // carry over is irrelevant
    return 1.0;
  }

  // handle regular points

  carry_forward *= (2.0*degree_f-1.0) / (2.0*degree_f);
  p_nm[degree] = sqrt(carry_forward);
  T temp = 0.0;

  for( int m = (int) (degree-1); m >= 0; --m) {
    p_nm[m] = -(sin_theta * temp + 2 * (m+1) * cos_theta * p_nm[m+1]);

    T m_f = (T) m+1;
    T scale2 = sqrt((degree_f + m_f) * (degree - m_f + 1));
    p_nm[m] /= scale2;

    temp = sin_theta * p_nm[m+1] * scale2;
  }

  T factor = (0 == (degree % 2)) ? 1.0 : -1.0;
  p_nm[0] *= factor;
  factor *= -sqrt(2);
  T f3 = 1.0;
  for( unsigned int m = 1; m < degree+1; ++m) {
    p_nm[m] *= sin_theta * f3 * factor; 
    f3 *= sin_theta;
    factor *= -1.0;
  }

  return carry_forward; 
}

} // end namespace Legendre

#endif // #ifndef _LEGENDRE_H_
